# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
NEW VERSION This module contains the implementation of all the rewrite rules on ZX-diagrams in PyZX.

Each rewrite rule consists of two methods: a matcher and a rewriter.
The matcher finds as many non-overlapping places where the rewrite rule can be applied.
The rewriter takes in a list of matches, and performs the necessary changes on the graph to implement the rewrite.

Each match function takes as input a Graph instance,
and an optional "filter function" that tells the matcher to only consider
the vertices or edges that the filter function accepts.
It outputs a list of "match" objects. What these objects look like differs
per rewrite rule.

The rewrite function takes as input a Graph instance and a list of match objects
of the appropriate type. It outputs a 4-tuple
(edges to add, vertices to remove, edges to remove, isolated vertices check).
The first of these should be fed to :meth:`~pyzx.graph.base.BaseGraph.add_edge_table`,
while the second and third should be fed to
:meth:`~graph.base.BaseGraph.remove_vertices` and :meth:`~pyzx.graph.base.BaseGraph.remove_edges`.
The last parameter is a Boolean that when true means that the rewrite rule can introduce
isolated vertices that should be removed by
:meth:`~pyzx.graph.base.BaseGraph.remove_isolated_vertices`.

Dealing with this output is done using either :func:`apply_rule` or :func:`pyzx.simplify.simp`.

Warning:
    There is no guarantee that the matcher does not affect the graph, and currently some matchers
    do in fact change the graph. Similarly, the rewrite function also changes the graph other
    than through the output it generates (for instance by adding vertices or changes phases).

"""

from typing import Tuple, List, Dict, Set, FrozenSet
from typing import Any, Callable, TypeVar, Optional, Union
from typing_extensions import Literal

from collections import Counter
from fractions import Fraction
import itertools

import numpy as np

from pyzx_param.graph.multigraph import Multigraph

from pyzx_param.utils import VertexType, EdgeType, get_w_partner, get_z_box_label, set_z_box_label, toggle_edge, vertex_is_w, vertex_is_zx, vertex_is_zx_like, FloatInt, FractionLike, get_w_io, vertex_is_z_like
from pyzx_param.graph.base import BaseGraph, VT, ET
from pyzx_param.symbolic import Poly

RewriteOutputType = Tuple[Dict[Tuple[VT,VT],List[int]], List[VT], List[ET], bool]
MatchObject = TypeVar('MatchObject')

def apply_rule(
        g: BaseGraph[VT,ET],
        rewrite: Callable[[BaseGraph[VT,ET], List[MatchObject]],RewriteOutputType[VT,ET]],
        m: List[MatchObject],
        check_isolated_vertices:bool=True
        ) -> None:
    etab, rem_verts, rem_edges, check_isolated_vertices = rewrite(g, m)

    g.add_edge_table(etab)
    g.remove_edges(rem_edges)
    g.remove_vertices(rem_verts)
    if check_isolated_vertices: g.remove_isolated_vertices()


MatchBialgType = Tuple[VT,VT,List[VT],List[VT]]

def match_bialg(g: BaseGraph[VT,ET]) -> List[MatchBialgType[VT]]:
    """Does the same as :func:`match_bialg_parallel` but with ``num=1``."""
    return match_bialg_parallel(g, num=1)


#TODO: make it be hadamard edge aware
def match_bialg_parallel(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[ET],bool]]=None,
        num: int=-1
        ) -> List[MatchBialgType[VT]]:
    """Finds noninteracting matchings of the bialgebra rule.

    :param g: An instance of a ZX-graph.
    :param matchf: An optional filtering function for candidate edge, should
       return True if a edge should considered as a match. Passing None will
       consider all edges.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :rtype: List of 4-tuples ``(v1, v2, neighbors_of_v1,neighbors_of_v2)``
    """
    if matchf is not None:
        candidates = set([e for e in g.edges() if matchf(e)])
    else:
        candidates = set(g.edge_set())
    phases = g.phases()
    types = g.types()

    i = 0
    m: List[MatchBialgType[VT]] = []
    while (num == -1 or i < num) and len(candidates) > 0:
        v0, v1 = g.edge_st(candidates.pop())
        if g.is_ground(v0) or g.is_ground(v1):
            continue
        v0t = types[v0]
        v1t = types[v1]
        v0p = phases[v0]
        v1p = phases[v1]
        if (v0p == 0 and v1p == 0 and
        ((v0t == VertexType.Z and v1t == VertexType.X) or (v0t == VertexType.X and v1t == VertexType.Z))):
            v0n = [n for n in g.neighbors(v0) if not n == v1]
            v1n = [n for n in g.neighbors(v1) if not n == v0]
            if (all([types[n] == v1t and phases[n] == 0 for n in v0n]) and # all neighbors of v0 are of the same type as v1
                all([types[n] == v0t and phases[n] == 0 for n in v1n]) and # all neighbors of v1 are of the same type as v0
                g.num_edges(v0, v1) == 1 and # there is exactly one edge between v0 and v1
                g.num_edges(v0, v0) == 0 and # there are no self-loops on v0
                g.num_edges(v1, v1) == 0): # there are no self-loops on v1
                i += 1
                for vn in [v0n, v1n]:
                    for v in vn:
                        for c in g.incident_edges(v):
                            if c in candidates:
                                candidates.discard(c)
                m.append((v0,v1,v0n,v1n))
    return m


def bialg(g: BaseGraph[VT,ET], matches: List[MatchBialgType[VT]]) -> RewriteOutputType[VT,ET]:
    """Performs a certain type of bialgebra rewrite given matchings supplied by
    ``match_bialg(_parallel)``."""
    rem_verts: List[VT] = []
    etab: Dict[Tuple[VT,VT], List[int]] = dict()
    for m in matches:
        rem_verts.append(m[0])
        rem_verts.append(m[1])
        es = [(i,j) for i in m[2] for j in m[3]]
        for e in es:
            if e in etab: etab[e][0] += 1
            else: etab[e] = [1,0]

    return (etab, rem_verts, [], True)

MatchSpiderType = Tuple[VT,VT]

def match_spider(g: BaseGraph[VT,ET]) -> List[MatchSpiderType[VT]]:
    """Does the same as :func:`match_spider_parallel` but with ``num=1``."""
    return match_spider_parallel(g, num=1)

def match_spider_parallel(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[VT],bool]]=None,
        num:int=-1
        ) -> List[MatchSpiderType[VT]]:
    """Finds non-interacting matchings of the spider fusion rule.

    :param g: An instance of a ZX-graph.
    :param matchf: An optional filtering function for candidate edge, should
       return True if the edge should be considered for matchings. Passing None will
       consider all edges.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :rtype: List of 2-tuples ``(v1, v2)``
    """
    if matchf is not None: candidates = set([v for v in g.vertices() if matchf(v)])
    else: candidates = g.vertex_set()
    candidates = set(candidates)
    types = g.types()
    m: List[MatchSpiderType[VT]] = []
    i = 0
    used: Set[VT] = set()
    for v in list(candidates):
        if v in used:
            continue
        for e in g.incident_edges(v):
            if g.edge_type(e) != EdgeType.SIMPLE:
                continue
            v0, v1 = g.edge_st(e)
            if v0 == v1 or v0 in used or v1 in used or v0 not in candidates or v1 not in candidates:
                continue
            v0t, v1t = types[v0], types[v1]
            if (v0t == v1t and vertex_is_zx(v0t)) or (vertex_is_z_like(v0t) and vertex_is_z_like(v1t)):
                i += 1
                used.update([v0, v1, *g.neighbors(v0), *g.neighbors(v1)])
                m.append((v0, v1))
                if num != -1 and i >= num:
                    return m
        if num != -1 and i >= num:
            break
    return m


def spider(g: BaseGraph[VT,ET], matches: List[MatchSpiderType[VT]]) -> RewriteOutputType[VT,ET]:
    '''Performs spider fusion given a list of matchings from ``match_spider(_parallel)``
    '''
    rem_verts: List[VT] = []
    etab: Dict[Tuple[VT,VT],List[int]] = dict()

    for m in matches:
        if g.row(m[0]) == 0:
            v0, v1 = m[1], m[0]
        else:
            v0, v1 = m[0], m[1]

        ground = g.is_ground(v0) or g.is_ground(v1)

        if ground:
            g.set_phase(v0, 0)
            g.set_ground(v0)
        elif g.type(v0) == VertexType.Z_BOX or g.type(v1) == VertexType.Z_BOX:
            if g.type(v0) == VertexType.Z:
                z_to_z_box(g, [v0])
            if g.type(v1) == VertexType.Z:
                z_to_z_box(g, [v1])
            g.set_phase(v0, 0)
            set_z_box_label(g, v0, get_z_box_label(g, v0) * get_z_box_label(g, v1))
        else:
            g.add_to_phase(v0, g.phase(v1), g.get_params(v1))

        if g.track_phases:
            g.fuse_phases(v0,v1)

        # always delete the second vertex in the match
        rem_verts.append(v1)

        # edges from the second vertex are transferred to the first
        for e in g.incident_edges(v1):
            source, target = g.edge_st(e)
            if source != v1:
                other_vertex = source
            elif target != v1:
                other_vertex = target
            else: #self-loop
                other_vertex = v0
            new_e = (v0, other_vertex)
            if new_e not in etab: etab[new_e] = [0,0]
            etab[new_e][g.edge_type(e)-1] += 1
        etab[(v0,v0)][0] = 0 # remove simple edge loops
    return (etab, rem_verts, [], True)

def unspider(g: BaseGraph[VT,ET], m: List[Any], qubit:FloatInt=-1, row:FloatInt=-1) -> VT:
    """Undoes a single spider fusion, given a match ``m``. A match is a list with 3
    elements given by::

      m[0] : a vertex to unspider
      m[1] : the neighbors of the new node, which should be a subset of the
             neighbors of m[0]
      m[2] : the phase of the new node. If omitted, the new node gets all of the phase of m[0]

    Returns the index of the new node. Optional parameters ``qubit`` and ``row`` can be used
    to position the new node. If they are omitted, they are set as the same as the old node.
    """
    u = m[0]
    v = g.add_vertex(ty=g.type(u))
    u_is_ground = g.is_ground(u)
    g.set_qubit(v, qubit if qubit != -1 else g.qubit(u))
    g.set_row(v, row if row != -1 else g.row(u))

    g.add_edge((u, v))
    for n in m[1]:
        e = g.edge(u,n)
        g.add_edge((v,n), edgetype=g.edge_type(e))
        g.remove_edge(e)
    if len(m) >= 3:
        g.add_to_phase(v, m[2])
        if not u_is_ground:
            g.add_to_phase(u, Fraction(0) - m[2])
    else:
        g.set_phase(v, g.phase(u))
        g.set_phase(u, 0)
    return v

def match_z_to_z_box(g: BaseGraph[VT,ET]) -> List[VT]:
    """Does the same as :func:`match_z_to_z_box_parallel` but with ``num=1``."""
    return match_z_to_z_box_parallel(g, num=1)

def match_z_to_z_box_parallel(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[VT],bool]]=None,
        num:int=-1
        ) -> List[VT]:
    """Finds all vertices that can be converted to Z-boxes."""
    if matchf is not None: candidates = set([v for v in g.vertices() if matchf(v)])
    else: candidates = g.vertex_set()
    types = g.types()
    phases = g.phases()
    m: List[VT] = []
    for v in candidates:
        if types[v] == VertexType.Z and not isinstance(phases[v],Poly):
            if num == 0: break
            m.append(v)
            num -= 1
    return m

def z_to_z_box(g: BaseGraph[VT,ET], matches: List[VT]) -> RewriteOutputType[VT,ET]:
    """Converts a Z vertex to a Z-box."""
    for v in matches:
        g.set_type(v, VertexType.Z_BOX)
        phase = g.phase(v)
        assert not isinstance(phase, Poly)
        label = np.round(np.e**(1j * np.pi * phase), 8)
        set_z_box_label(g, v, label)
        g.set_phase(v, 0)
    return ({}, [], [], True)

MatchWType = Tuple[VT,VT]

def match_w_fusion(g: BaseGraph[VT,ET]) -> List[MatchWType[VT]]:
    """Does the same as :func:`match_w_fusion_parallel` but with ``num=1``."""
    return match_w_fusion_parallel(g, num=1)

def match_w_fusion_parallel(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[ET],bool]]=None,
        num:int=-1
        ) -> List[MatchWType[VT]]:
    """Finds non-interacting matchings of the W fusion rule.

    :param g: An instance of a ZX-graph.
    :param matchf: An optional filtering function for candidate edge, should
       return True if the edge should be considered for matchings. Passing None will
       consider all edges.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :rtype: List of 2-tuples ``(v1, v2)``
    """
    if matchf is not None: candidates = set([e for e in g.edges() if matchf(e)])
    else: candidates = g.edge_set()
    types = g.types()

    i = 0
    m: List[MatchWType[VT]] = []
    while (num == -1 or i < num) and len(candidates) > 0:
        e = candidates.pop()
        if g.edge_type(e) != EdgeType.SIMPLE: continue
        v0, v1 = g.edge_st(e)
        v0t = types[v0]
        v1t = types[v1]
        if vertex_is_w(v0t) and vertex_is_w(v1t):
            i += 1
            candidates_to_remove: List[VT] = []
            candidates_to_remove.extend(list(g.neighbors(v0)))
            candidates_to_remove.extend(list(g.neighbors(v1)))
            candidates_to_remove.extend(list(g.neighbors(get_w_partner(g, v0))))
            candidates_to_remove.extend(list(g.neighbors(get_w_partner(g, v1))))
            for v in candidates_to_remove:
                for c in g.incident_edges(v): candidates.discard(c)
            m.append((v0,v1))
    return m

def w_fusion(g: BaseGraph[VT,ET], matches: List[MatchSpiderType[VT]]) -> RewriteOutputType[VT,ET]:
    '''Performs W fusion given a list of matchings from ``match_w_fusion(_parallel)``
    '''
    rem_verts: List[VT] = []
    etab: Dict[Tuple[VT,VT],List[int]] = dict()

    for v0, v1 in matches:
        v0_in, v0_out = get_w_io(g, v0)
        v1_in, v1_out = get_w_io(g, v1)
        if not g.connected(v0_out, v1_in):
            v0_in, v1_in = v1_in, v0_in
            v0_out, v1_out = v1_out, v0_out
        # always delete the second vertex in the match
        rem_verts.extend([v1_in, v1_out])

        # edges from the second vertex are transferred to the first
        for w in g.neighbors(v1_out):
            if w == v1_in or w == v0_in:
                continue
            if w == v1_out:
                w = v0_out
            e = (v0_out, w)
            if e not in etab: etab[e] = [0,0]
            etab[e][g.edge_type(g.edge(v1_out, w)) - 1] += 1
    return (etab, rem_verts, [], True)


MatchPivotType = Tuple[Tuple[VT,VT],Tuple[List[VT],List[VT]]]

def match_pivot(g: BaseGraph[VT,ET]) -> List[MatchPivotType[VT]]:
    """Does the same as :func:`match_pivot_parallel` but with ``num=1``."""
    return match_pivot_parallel(g, num=1, check_edge_types=True)


def match_pivot_parallel(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[ET],bool]]=None,
        num:int=-1,
        check_edge_types:bool=True
        ) -> List[MatchPivotType[VT]]:
    """Finds non-interacting matchings of the pivot rule.

    :param g: An instance of a ZX-graph.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :param check_edge_types: Whether the method has to check if all the edges involved
       are of the correct type (Hadamard edges).
    :param matchf: An optional filtering function for candidate edge, should
       return True if a edge should considered as a match. Passing None will
       consider all edges.
    :rtype: List of 4-tuples. See :func:`pivot` for the details.
    """
    if matchf is not None:
        candidates = set([e for e in g.edges() if matchf(e)])
    else:
        candidates = set(g.edge_set())
    types = g.types()
    phases = g.phases()

    i = 0
    m: List[MatchPivotType[VT]] = []
    while (num == -1 or i < num) and len(candidates) > 0:
        e = candidates.pop()
        if check_edge_types and g.edge_type(e) != EdgeType.HADAMARD: continue
        v0, v1 = g.edge_st(e)

        if not (types[v0] == VertexType.Z and types[v1] == VertexType.Z): continue

        v0a = phases[v0]
        v1a = phases[v1]
        if not ((v0a in (0,1)) and (v1a in (0,1))): continue
        if g.is_ground(v0) or g.is_ground(v1):
            continue

        invalid_edge = False

        v0n = list(g.neighbors(v0))
        v0b: List[VT] = []
        for n in v0n:
            if len(list(g.edges(v0,n))) != 1:
                invalid_edge = True
                break
            et = g.edge_type(g.edge(v0,n))
            if types[n] == VertexType.Z and et == EdgeType.HADAMARD: pass
            elif types[n] == VertexType.BOUNDARY: v0b.append(n)
            else:
                invalid_edge = True
                break

        if invalid_edge: continue

        v1n = list(g.neighbors(v1))
        v1b: List[VT] = []
        for n in v1n:
            et = g.edge_type(g.edge(v1,n))
            if types[n] == VertexType.Z and et == EdgeType.HADAMARD: pass
            elif types[n] == VertexType.BOUNDARY: v1b.append(n)
            else:
                invalid_edge = True
                break

        if invalid_edge: continue
        if len(v0b) + len(v1b) > 1: continue

        i += 1
        for vn in [v0n, v1n]:
            for v in vn:
                for c in g.incident_edges(v):
                    if c in candidates:
                        candidates.discard(c)
        b0 = list(v0b)
        b1 = list(v1b)
        m.append(((v0,v1),(b0,b1)))
    return m

def match_pivot_gadget(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[ET],bool]]=None,
        num:int=-1) -> List[MatchPivotType[VT]]:
    """Like :func:`match_pivot_parallel`, but except for pairings of
    Pauli vertices, it looks for a pair of an interior Pauli vertex and an
    interior non-Clifford vertex in order to gadgetize the non-Clifford vertex."""
    if matchf is not None:
        candidates = set([e for e in g.edges() if matchf(e)])
    else:
        candidates = set(g.edge_set())
    types = g.types()
    phases = g.phases()
    rs = g.rows()

    edge_list: List[Tuple[VT,VT]] = []
    i = 0
    m: List[MatchPivotType[VT]] = []
    while (num == -1 or i < num) and len(candidates) > 0:
        e = candidates.pop()
        v0, v1 = g.edge_st(e)

        if not (types[v0] == VertexType.Z and types[v1] == VertexType.Z): continue

        v0a = phases[v0]
        v1a = phases[v1]
        v0ps = g.get_params(v0)
        v1ps = g.get_params(v1)

        if v0a not in (0,1):
            if v1a in (0,1):
                v0, v1 = v1, v0
                v0a, v1a = v1a, v0a
                v0ps, v1ps = v1ps, v0ps
            else: continue
        elif v1a in (0,1): continue
        # Now v0 has a Pauli phase and v1 has a non-Pauli phase

        if g.is_ground(v0):
            continue

        v0n = list(g.neighbors(v0))
        v1n = list(g.neighbors(v1))
        if len(v1n) == 1: continue # It is a phase gadget
        bad_match = False
        discard_edges: List[ET] = []
        for i,l in enumerate((v0n, v1n)):
            for n in l:
                if types[n] != VertexType.Z:
                    bad_match = True
                    break
                ne = list(g.incident_edges(n))
                if i==0 and len(ne) == 1 and not (e == ne[0]): # v0 is a phase gadget
                    bad_match = True
                    break
                discard_edges.extend(ne)
            if bad_match: break
        if bad_match: continue

        if any(types[w]!=VertexType.Z for w in v0n): continue
        if any(types[w]!=VertexType.Z for w in v1n): continue
        # Both v0 and v1 are interior

        v = g.add_vertex(VertexType.Z,-2,rs[v0],v1a)
        g.set_params(v, v1ps)
        g.set_phase(v1, 0)
        g.set_qubit(v0,-1)
        g.update_phase_index(v1,v)
        edge_list.append((v,v1))

        m.append(((v0,v1),([],[v])))
        i += 1
        for c in discard_edges:
            if c in candidates:
                candidates.remove(c)
    g.add_edges(edge_list,EdgeType.SIMPLE)
    return m


def match_pivot_boundary(
        g: BaseGraph[VT,ET],
        matchf:Optional[Callable[[VT],bool]]=None,
        num:int=-1) -> List[MatchPivotType[VT]]:
    """Like :func:`match_pivot_parallel`, but except for pairings of
    Pauli vertices, it looks for a pair of an interior Pauli vertex and a
    boundary non-Pauli Clifford vertex in order to gadgetize the non-Pauli vertex."""
    if matchf is not None: candidates = set([v for v in g.vertices() if matchf(v)])
    else: candidates = g.vertex_set()
    types = g.types()
    phases = g.phases()
    rs = g.rows()

    edge_list: List[Tuple[VT,VT]] = []
    consumed_vertices : Set[VT] = set()
    i = 0
    m: List[MatchPivotType[VT]] = []
    inputs = g.inputs()
    while (num == -1 or i < num) and len(candidates) > 0:
        v = candidates.pop()
        if types[v] != VertexType.Z or phases[v] not in (0,1) or g.is_ground(v):
            continue

        good_vert = True
        w = None
        bound = None
        for n in g.neighbors(v):
            if types[n] != VertexType.Z:
                good_vert = False
                break
            if len(g.neighbors(n)) == 1: # v is a phase gadget
                good_vert = False
                break
            if n in consumed_vertices:
                good_vert = False
                break
            if g.is_ground(n) in consumed_vertices:
                good_vert = False
                break
            boundaries: List[VT] = []
            wrong_match = False
            for b in g.neighbors(n):
                if types[b] == VertexType.BOUNDARY:
                    boundaries.append(b)
                elif types[b] != VertexType.Z:
                    wrong_match = True
            if len(boundaries) != 1 or wrong_match: # n is not on the boundary,
                continue             # has too many boundaries or has neighbors of wrong type
            if phases[n] and hasattr(phases[n], 'denominator') and phases[n].denominator == 2:
                w = n
                bound = boundaries[0]
            if not w:
                w = n
                bound = boundaries[0]
        if not good_vert or w is None: continue
        if bound in inputs: mod = 0.5
        else: mod = -0.5
        v1 = g.add_vertex(VertexType.Z,-2,rs[w]+mod,phases[w], phaseVars=g.get_params(w))
        v2 = g.add_vertex(VertexType.Z,-1,rs[w]+mod,0)
        g.set_phase(w, 0)
        g.update_phase_index(w,v1)
        edge_list.append((w,v2))
        edge_list.append((v1,v2))
        for n in g.neighbors(v): consumed_vertices.add(n)
        for n in g.neighbors(w): consumed_vertices.add(n)
        assert bound is not None
        m.append(((v,w),([],[bound])))
        i += 1
        for n in g.neighbors(v): candidates.discard(n)
        for n in g.neighbors(w): candidates.discard(n)

    g.add_edges(edge_list, EdgeType.HADAMARD)
    return m

def pivot(g: BaseGraph[VT,ET], matches: List[MatchPivotType[VT]]) -> RewriteOutputType[VT,ET]:
    """Perform a pivoting rewrite, given a list of matches as returned by
    ``match_pivot(_parallel)``. A match is itself a list where:

    ``m[0][0]`` : first vertex in pivot.
    ``m[0][1]`` : second vertex in pivot.
    ``m[1][0]`` : list of zero or one boundaries adjacent to ``m[0]``.
    ``m[1][1]`` : list of zero or one boundaries adjacent to ``m[1]``.
    """
    rem_verts: List[VT] = []
    rem_edges: List[ET] = []
    etab: Dict[Tuple[VT,VT],List[int]] = dict()


    for m in matches:
        # compute:
        #  n[0] <- non-boundary neighbors of m[0] only
        #  n[1] <- non-boundary neighbors of m[1] only
        #  n[2] <- non-boundary neighbors of m[0] and m[1]
        g.update_phase_index(m[0][0],m[0][1])
        n = [set(g.neighbors(m[0][0])), set(g.neighbors(m[0][1]))]
        for i in range(2):
            n[i].remove(m[0][1-i])
            if len(m[1][i]) == 1: n[i].remove(m[1][i][0])
        n.append(n[0] & n[1])
        n[0] = n[0] - n[2]
        n[1] = n[1] - n[2]
        es = ([(s,t) for s in n[0] for t in n[1]] +
              [(s,t) for s in n[1] for t in n[2]] +
              [(s,t) for s in n[0] for t in n[2]])
        k0, k1, k2 = len(n[0]), len(n[1]), len(n[2])
        g.scalar.add_power(k0*k2 + k1*k2 + k0*k1)

        for v in n[2]:
            if not g.is_ground(v):
                g.add_to_phase(v, 1)

        p0 = g.phase(m[0][0])
        p1 = g.phase(m[0][1])
        ps0 = g.get_params(m[0][0])
        ps1 = g.get_params(m[0][1])
        if len(ps0) == 0 and len(ps1) == 0:
            if p0 and p1: # no params and both non-zero
                g.scalar.add_phase(Fraction(1))
        elif (p0 in (0, 1)) and (p1 in (0, 1)): # one or more has params, and both have 0 or 1 const
            psA = ps0
            psB = ps1
            if (p0 == 1): psA = psA.symmetric_difference({"1"})
            if (p1 == 1): psB = psB.symmetric_difference({"1"})
            g.scalar.add_phase_vars_pi_pair(psA, psB)
        elif (p0 in (0, 1)) or (p1 in (0, 1)): # only one has a 0 or 1 const
            if (p0 == 0): g.scalar.add_phase_vars_pi(ps0)
            elif (p0 == 1): g.scalar.add_phase_vars_pi(ps0.symmetric_difference({"1"}))
            if (p1 == 0): g.scalar.add_phase_vars_pi(ps1)
            elif (p1 == 1): g.scalar.add_phase_vars_pi(ps1.symmetric_difference({"1"}))
        else: # Neither has 0 or 1 const, hence both must be non-zero
            g.scalar.add_phase(Fraction(1))
        
        if not m[1][0] and not m[1][1]:
            g.scalar.add_power(-(k0+k1+2*k2-1))
        elif not m[1][0]:
            g.scalar.add_power(-(k1+k2))
        else: g.scalar.add_power(-(k0+k2))

        for i in 0, 1:
            # if m[i] has a phase, it will get copied on to the neighbors of m[1-i]:
            a = g.phase(m[0][i]) # type: ignore
            ps = g.get_params(m[0][i])
            for v in n[1-i]:
                if not g.is_ground(v):
                    g.add_to_phase(v, a, ps)
            for v in n[2]:
                if not g.is_ground(v):
                    g.add_to_phase(v, a, ps)

            if not m[1][i]:
                # if there is no boundary, the other vertex is destroyed
                rem_verts.append(m[0][1-i])
            else:
                # if there is a boundary, toggle whether it is an h-edge or a normal edge
                # and point it at the other vertex
                e = (m[0][i], m[1][i][0])
                new_e = (m[0][1-i], m[1][i][0])
                ne,nhe = etab.get(new_e) or [0,0]
                if g.edge_type(g.edge(e[0],e[1])) == EdgeType.SIMPLE: nhe += 1
                elif g.edge_type(g.edge(e[0],e[1])) == EdgeType.HADAMARD: ne += 1
                etab[new_e] = [ne,nhe]
                rem_edges.append(g.edge(e[0], e[1]))


        for e in es:
            nhe = etab.get(e, (0,0))[1]
            etab[e] = [0,nhe+1]

    return (etab, rem_verts, rem_edges, True)

MatchLcompType = Tuple[VT,List[VT]]

def match_lcomp(g: BaseGraph[VT,ET]) -> List[MatchLcompType[VT]]:
    """Same as :func:`match_lcomp_parallel`, but with ``num=1``"""
    return match_lcomp_parallel(g, num=1, check_edge_types=True)

def match_lcomp_parallel(
        g: BaseGraph[VT,ET],
        vertexf:Optional[Callable[[VT],bool]]=None,
        num:int=-1,
        check_edge_types:bool=True
        ) -> List[MatchLcompType[VT]]:
    """Finds noninteracting matchings of the local complementation rule.

    :param g: An instance of a ZX-graph.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :param check_edge_types: Whether the method has to check if all the edges involved
       are of the correct type (Hadamard edges).
    :param vertexf: An optional filtering function for candidate vertices, should
       return True if a vertex should be considered as a match. Passing None will
       consider all vertices.
    :rtype: List of 2-tuples ``(vertex, neighbors)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()
    phases = g.phases()

    i = 0
    m: List[MatchLcompType[VT]] = []
    while (num == -1 or i < num) and len(candidates) > 0:
        v = candidates.pop()
        vt = types[v]
        va = g.phase(v)

        if vt != VertexType.Z: continue
        if not (va == Fraction(1,2) or va == Fraction(3,2)): continue

        if g.is_ground(v):
            continue

        if check_edge_types and not (
            all(g.edge_type(e) == EdgeType.HADAMARD for e in g.incident_edges(v))
            ): continue

        vn = list(g.neighbors(v))

        if not all(types[n] == VertexType.Z for n in vn): continue

        for n in vn: candidates.discard(n)
        m.append((v,vn))
    return m

def lcomp(g: BaseGraph[VT,ET], matches: List[MatchLcompType[VT]]) -> RewriteOutputType[VT,ET]:
    """Performs a local complementation based rewrite rule on the given graph with the
    given ``matches`` returned from ``match_lcomp(_parallel)``. See "Graph Theoretic
    Simplification of Quantum Circuits using the ZX calculus" (arXiv:1902.03178)
    for more details on the rewrite"""
    etab: Dict[Tuple[VT,VT],List[int]] = dict()
    rem: List[VT] = []
    for m in matches:
        a = g.phase(m[0])
        ps = g.get_params(m[0])
        rem.append(m[0])
        assert isinstance(a,Fraction)  # For mypy
        if a.numerator == 1: # if a = 1/2
            g.scalar.add_phase(Fraction(1,4))
            g.scalar.add_phase_vars_halfpi(ps, 3)    # Add (a+b+...)*(3/2)
        else:                # if a = 3/2
            g.scalar.add_phase(Fraction(7,4))
            g.scalar.add_phase_vars_halfpi(ps, 1)    # Add (a+b+...)*(1/2)
        n = len(m[1])
        g.scalar.add_power((n-2)*(n-1)//2)
        for i in range(n):
            if not g.is_ground(m[1][i]):
                g.add_to_phase(m[1][i], -a, ps)
            for j in range(i+1, n):
                e = (m[1][i],m[1][j])
                he = etab.get(e, [0,0])[1]
                etab[e] = [0, he+1]

    return (etab, rem, [], True)

MatchIdType = Tuple[VT,VT,VT,EdgeType]

def match_ids(g: BaseGraph[VT,ET]) -> List[MatchIdType[VT]]:
    """Finds a single identity node. See :func:`match_ids_parallel`."""
    return match_ids_parallel(g, num=1)

def match_ids_parallel(
        g: BaseGraph[VT,ET],
        vertexf:Optional[Callable[[VT],bool]]=None,
        num:int=-1
        ) -> List[MatchIdType[VT]]:
    """Finds non-interacting identity vertices.

    :param g: An instance of a ZX-graph.
    :param num: Maximal amount of matchings to find. If -1 (the default)
       tries to find as many as possible.
    :param vertexf: An optional filtering function for candidate vertices, should
       return True if a vertex should be considered as a match. Passing None will
       consider all vertices.
    :rtype: List of 4-tuples ``(identity_vertex, neighbor1, neighbor2, edge_type)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()
    phases = g.phases()

    i = 0
    m:List[MatchIdType[VT]] = []

    while (num == -1 or i < num) and len(candidates) > 0:
        v = candidates.pop()
        if phases[v] != 0 or not vertex_is_zx(types[v]) or g.is_ground(v) or g.vertex_degree(v) != 2 or len(g.get_params(v)) > 0:
            continue
        if len(g.incident_edges(v)) != 2: continue
        neigh = g.neighbors(v)
        if len(neigh) == 2:
            v0, v1 = neigh
        else:
            v0, v1 = list(neigh)[0], list(neigh)[0]
        if (g.is_ground(v0) and types[v1] == VertexType.BOUNDARY or
                g.is_ground(v1) and types[v0] == VertexType.BOUNDARY):
            # Do not put ground spiders on the boundary
            continue
        candidates.discard(v0)
        candidates.discard(v1)
        if g.edge_type(g.edge(v,v0)) != g.edge_type(g.edge(v,v1)): #exactly one of them is a hadamard edge
            m.append((v,v0,v1,EdgeType.HADAMARD))
        else: m.append((v,v0,v1,EdgeType.SIMPLE))
        i += 1
    return m

def remove_ids(g: BaseGraph[VT,ET], matches: List[MatchIdType[VT]]) -> RewriteOutputType[VT,ET]:
    """Given the output of ``match_ids(_parallel)``, returns a list of edges to add,
    and vertices to remove."""
    etab : Dict[Tuple[VT,VT],List[int]] = dict()
    rem: List[VT] = []
    for v,v0,v1,et in matches:
        rem.append(v)
        e = (v0,v1)
        if not e in etab: etab[e] = [0,0]
        if et == EdgeType.SIMPLE: etab[e][0] += 1
        else: etab[e][1] += 1
    return (etab, rem, [], False)


MatchHopfType = Tuple[VT, VT, EdgeType]

def match_hopf(g: BaseGraph[VT,ET], vertexf:Optional[Callable[[VT],bool]]=None) -> List[MatchHopfType[VT]]:
    """Finds parallel edges between spiders that can be removed.
    :param g: An instance of a ZX-graph.
    :param vertexf: An optional filtering function for candidate vertices, should
    return True if a vertex should be considered as a match. Passing None will
    consider all vertices.
    :rtype: List of 2-tuples ``(vertex, neighbors)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    m:List[MatchHopfType] = []

    for v in candidates:
        if not vertex_is_zx_like(types[v]): continue
        for w in g.neighbors(v):
            if not w in candidates or not vertex_is_zx_like(types[w]): continue

            # If the number of edges between v and w is greater than 1, we can remove edges
            ns = g.num_edges(v, w, EdgeType.SIMPLE)
            nh = g.num_edges(v, w, EdgeType.HADAMARD)
            v_is_z = vertex_is_z_like(types[v])
            w_is_z = vertex_is_z_like(types[w])
            if v_is_z == w_is_z: # Both Z-like or X-like
                if nh > 1:
                    if (w,v,EdgeType.HADAMARD) not in m:
                        m.append((v,w,EdgeType.HADAMARD))
            else: # One is Z-like, the other is X-like
                if ns > 1:
                    if (w,v,EdgeType.SIMPLE) not in m:
                        m.append((v,w,EdgeType.SIMPLE))
    return m

def hopf(g: BaseGraph[VT,ET], matches: List[MatchHopfType[VT]]) -> RewriteOutputType[VT,ET]:
    """Performs a Hopf rule rewrite on the given graph with the
    given ``matches`` returned from ``match_hopf``. Removes all parallel edges of the given type
    A match is itself a list where:

    ``m[0]`` : first vertex in Hopf.
    ``m[1]`` : second vertex in Hopf.
    ``m[2]`` : edge type of the edge to remove.
    """
    etab: Dict[Tuple[VT,VT],List[int]] = dict()
    rem_edges: List[ET] = []
    for v,w,et in matches:
        n = g.num_edges(v,w,et)
        parity = n % 2
        rem_edges.extend([g.edge(v, w, et)]*(n - parity))
        g.scalar.add_power(-(n-parity))

    return (etab, [], rem_edges, False)
    
MatchSelfLoopType = Tuple[VT, int, int]
def match_self_loop(g: BaseGraph[VT,ET], vertexf:Optional[Callable[[VT],bool]]=None) -> List[MatchSelfLoopType[VT]]:
    """Finds self-loops on vertices that can be removed.
    :param g: An instance of a ZX-graph.
    :param vertexf: An optional filtering function for candidate vertices, should
    return True if a vertex should be considered as a match. Passing None will
    consider all vertices.
    :rtype: List of 2-tuples ``(vertex, neighbors)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    m:List[MatchSelfLoopType] = []
    
    for v in candidates:
        if not vertex_is_zx_like(types[v]): continue

        ns = g.num_edges(v, v, EdgeType.SIMPLE)
        nh = g.num_edges(v, v, EdgeType.HADAMARD)
        if ns == 0 and nh == 0: continue
        m.append((v,ns,nh))
    return m

def remove_self_loops(g: BaseGraph[VT,ET], matches: List[MatchSelfLoopType[VT]]) -> RewriteOutputType[VT,ET]:
    """Performs a self-loop removal rewrite on the given graph with the
    given ``matches`` returned from ``match_self_loop``. Removes all self-loops of the given type
    A match is itself a list where:

    ``m[0]`` : vertex in self-loop.
    ``m[1]`` : number of simple edges to remove.
    ``m[2]`` : number of hadamard edges to remove.
    """
    etab: Dict[Tuple[VT,VT],List[int]] = dict()
    rem_edges: List[ET] = []
    for v,ns,nh in matches:
        rem_edges.extend([g.edge(v, v, EdgeType.SIMPLE)]*ns)
        rem_edges.extend([g.edge(v, v, EdgeType.HADAMARD)]*nh)
        g.scalar.add_power(-nh)
        if nh % 2 == 1: # A Hadamard self-loop gives a phase of pi
            g.add_to_phase(v, Fraction(1,1))

    return (etab, [], rem_edges, False)

MatchGadgetType = Tuple[VT,VT,FractionLike,List[VT],List[VT]]

def match_phase_gadgets(g: BaseGraph[VT,ET],vertexf:Optional[Callable[[VT],bool]]=None) -> List[MatchGadgetType[VT]]:
    """Determines which phase gadgets act on the same vertices, so that they can be fused together.

    :param g: An instance of a ZX-graph.
    :rtype: List of 5-tuples ``(axel,leaf, total combined phase, other axels with same targets, other leafs)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()

    phases = g.phases()

    parities: Dict[FrozenSet[VT], List[VT]] = dict()
    gadgets: Dict[VT,VT] = dict()
    inputs = g.inputs()
    outputs = g.outputs()
    # First we find all the phase-gadgets, and the list of vertices they act on
    for v in candidates:
        if isinstance(phases[v], Poly):
            non_clifford = True
        else:
            non_clifford = phases[v] != 0 and getattr(phases[v], 'denominator', 1) > 2
        if non_clifford and len(list(g.neighbors(v)))==1:
            n = list(g.neighbors(v))[0]
            if phases[n] not in (0,1): continue # Not a real phase gadget (happens for scalar diagrams)
            if n in gadgets: continue # Not a real phase gadget (happens for scalar diagrams)
            if n in inputs or n in outputs: continue # Not a real phase gadget (happens for non-unitary diagrams)
            if len(g.get_params(v)) > 0: continue # Not supported for symbolic params
            gadgets[n] = v
            par = frozenset(set(g.neighbors(n)).difference({v}))
            if par == frozenset(): continue # Not a real phase gadget if it acts on nothing
            if par in parities: parities[par].append(n)
            else: parities[par] = [n]

    m: List[MatchGadgetType[VT]] = []
    for par, gad in parities.items():
        if len(gad) == 1:
            n = gad[0]
            v = gadgets[n]
            if phases[n] != 0: # If the phase of the axel vertex is pi, we change the phase of the gadget
                g.scalar.add_phase(phases[v])
                g.phase_negate(v)
                m.append((v,n,-phases[v],[],[]))
        else:
            totphase = sum((1 if phases[n]==0 else -1)*phases[gadgets[n]] for n in gad)%2
            for n in gad:
                if phases[n] != 0:
                    g.scalar.add_phase(phases[gadgets[n]])
                    g.phase_negate(gadgets[n])
            g.scalar.add_power(-((len(par)-1)*(len(gad)-1)))
            n = gad.pop()
            v = gadgets[n]
            m.append((v,n,totphase, gad, [gadgets[n] for n in gad]))
    return m

def merge_phase_gadgets(g: BaseGraph[VT,ET], matches: List[MatchGadgetType[VT]]) -> RewriteOutputType[VT,ET]:
    """Given the output of :func:``match_phase_gadgets``, removes phase gadgets that act on the same set of targets."""
    rem: List[VT] = []
    for v, n, phase, othergadgets, othertargets in matches:
        g.set_phase(v, phase)
        g.set_phase(n, 0)
        rem.extend(othergadgets)
        rem.extend(othertargets)
        for w in othertargets:
            g.fuse_phases(v,w)
            if g.merge_vdata is not None:
                g.merge_vdata(v, w)
    return ({}, rem, [], False)

MatchSupplementarityType = Tuple[VT,VT,Literal[1,2],FrozenSet[VT]]

def match_supplementarity(g: BaseGraph[VT,ET], vertexf:Optional[Callable[[VT],bool]]=None) -> List[MatchSupplementarityType[VT]]:
    """Finds pairs of non-Clifford spiders that are connected to exactly the same set of vertices.

    :param g: An instance of a ZX-graph.
    :rtype: List of 4-tuples ``(vertex1, vertex2, type of supplementarity, neighbors)``.
    """
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    phases = g.phases()

    parities: Dict[FrozenSet[VT],List[VT]] = dict()
    m: List[MatchSupplementarityType[VT]] = []
    taken: Set[VT] = set()
    # First we find all the non-Clifford vertices and their list of neighbors
    while len(candidates) > 0:
        v = candidates.pop()
        if phases[v] == 0 or (not isinstance(phases[v], Poly) and phases[v].denominator <= 2): continue # Skip Clifford vertices
        if phases[v].denominator not in [1,2,4]:
            # Skip non-Clifford+T vertices
            continue
        neigh = set(g.neighbors(v))
        if not neigh.isdisjoint(taken): continue
        par = frozenset(neigh)
        if par in parities:
            for w in parities[par]:
                if (phases[v]-phases[w]) % 2 == 1 or (phases[v]+phases[w]) % 2 == 1:
                    m.append((v,w,1,par))
                    taken.update({v,w})
                    taken.update(neigh)
                    candidates.difference_update(neigh)
                    break
            else: parities[par].append(v)
            if v in taken: continue
        else: parities[par] = [v]
        for w in neigh:
            if phases[w] == 0 or (not isinstance(phases[w], Poly) and phases[w].denominator <= 2) or w in taken: continue
            diff = neigh.symmetric_difference(g.neighbors(w))
            if len(diff) == 2: # Perfect overlap
                if (phases[v] + phases[w]) % 2 == 0 or (phases[v] - phases[w]) % 2 == 1:
                    m.append((v,w,2,frozenset(neigh.difference({w}))))
                    taken.update({v,w})
                    taken.update(neigh)
                    candidates.difference_update(neigh)
                    break
    return m

def apply_supplementarity(
        g: BaseGraph[VT,ET],
        matches: List[MatchSupplementarityType[VT]]
        ) -> RewriteOutputType[VT,ET]:
    """Given the output of :func:``match_supplementarity``, removes non-Clifford spiders that act on the same set of targets trough supplementarity.
    
    Warning:
        Takes in a graph-like diagram.  
        """
    rem: List[VT] = []
    for v, w, t, neigh in matches:
        rem.append(v)
        rem.append(w)
        alpha = g.phase(v)
        beta = g.phase(w)
        g.scalar.add_power(-2*len(neigh))
        if t == 1: # v and w are not connected
            g.scalar.add_node(2*alpha+1)
            #if (alpha-beta)%2 == 1: # Standard supplementarity
            if (alpha+beta)%2 == 1: # Need negation on beta
                g.scalar.add_phase(-alpha + 1)
                for n in neigh:
                    g.add_to_phase(n,1)
        elif t == 2: # they are connected
            g.scalar.add_power(-1)
            g.scalar.add_node(2*alpha)
            #if (alpha-beta)%2 == 1: # Standard supplementarity
            if (alpha+beta)%2 == 0: # Need negation
                g.scalar.add_phase(-alpha)
                for n in neigh:
                    g.add_to_phase(n,1)
        else: raise Exception("Shouldn't happen")
    return ({}, rem, [], True)

MatchCopyType = Tuple[VT,VT,FractionLike,FractionLike,List[VT]]

def match_copy(
    g: BaseGraph[VT,ET],
    vertexf:Optional[Callable[[VT],bool]]=None
    ) -> List[MatchCopyType[VT]]:
    """Finds spiders with a 0 or pi phase that have a single neighbor,
    and copies them through. Assumes that all the spiders are green and maximally fused."""
    if vertexf is not None:
        candidates = set([v for v in g.vertices() if vertexf(v)])
    else:
        candidates = g.vertex_set()
    phases = g.phases()
    types = g.types()
    m: List[MatchCopyType[VT]] = []

    while len(candidates) > 0:
        v = candidates.pop()
        if phases[v] not in (0,1) or types[v] != VertexType.Z or g.vertex_degree(v) != 1: continue
        w = list(g.neighbors(v))[0]
        if types[w] != VertexType.Z: continue
        neigh = [n for n in g.neighbors(w) if n != v]
        m.append((v,w,phases[v],phases[w],neigh))
        candidates.discard(w)
        candidates.difference_update(neigh)

    return m

def apply_copy(g: BaseGraph[VT,ET], matches: List[MatchCopyType[VT]]) -> RewriteOutputType[VT,ET]:
    rem: List[VT] = []
    types = g.types()
    outputs = g.outputs()
    for v,w,a,alpha, neigh in matches:
        assert len(g.neighbors(v)) == 1
        assert g.phase(v) in (0,1)
        w = list(g.neighbors(v))[0]
        neigh = [n for n in g.neighbors(w) if n != v]
        a = g.phase(v)
        alpha = g.phase(w)

        rem.append(v)
        rem.append(w)
        g.scalar.add_power(-len(neigh)+1)
        if a: g.scalar.add_phase(alpha)
        for n in neigh:
            if types[n] == VertexType.BOUNDARY:
                r = g.row(n) - 1 if n in outputs else g.row(n)+1
                u = g.add_vertex(VertexType.Z, g.qubit(n), r, a)
                e = g.edge(w,n)
                et = g.edge_type(e)
                g.add_edge((n,u), toggle_edge(et))
            else:
                g.add_to_phase(n, a)
        g.remove_vertices(rem)
        rem = []
    # return ({}, rem, [], True)
    return ({}, [], [], True)

MatchPhasePolyType = Tuple[List[VT], Dict[FrozenSet[VT],Union[VT,Tuple[VT,VT]]]]


def match_gadgets_phasepoly(g: BaseGraph[VT,ET]) -> List[MatchPhasePolyType[VT]]:
    """Finds groups of phase-gadgets that act on the same set of 4 vertices in order to apply a rewrite based on
    rule R_13 of the paper *A Finite Presentation of CNOT-Dihedral Operators*."""
    targets: Dict[VT,Set[FrozenSet[VT]]] = {}
    gadgets: Dict[FrozenSet[VT], Tuple[VT,VT]] = {}
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        if v not in inputs and v not in outputs and len(list(g.neighbors(v)))==1:
            if g.phase(v) != 0 and g.phase(v).denominator != 4: continue
            n = list(g.neighbors(v))[0]
            tgts = frozenset(set(g.neighbors(n)).difference({v}))
            if len(tgts)>4: continue
            gadgets[tgts] = (n,v)
            for t in tgts:
                if t in targets: targets[t].add(tgts)
                else: targets[t] = {tgts}
        if g.phase(v) != 0 and g.phase(v).denominator == 4:
            if v in targets: targets[v].add(frozenset([v]))
            else: targets[v] = {frozenset([v])}
    targets = {t:s for t,s in targets.items() if len(s)>1}
    matches: Dict[FrozenSet[VT], Set[FrozenSet[VT]]] = {}

    for v1,t1 in targets.items():
        s = t1.difference(frozenset([v1]))
        if len(s) == 1:
            c = s.pop()
            if any(len(targets[v2])==2 for v2 in c): continue
        s = t1.difference({frozenset({v1})})
        for c in [d for d in s if not any(d.issuperset(e) for e in s if e!=d)]:
            if not all(v2 in targets for v2 in c): continue
            if any(v2<v1 for v2 in c): continue
            a = set()
            for t in c: a.update([i for s in targets[t] for i in s if i in targets])
            for group in itertools.combinations(a.difference(c),4-len(c)):
                gr = list(group)+list(c)
                b: Set[FrozenSet[VT]] = set()
                for t in gr: b.update([s for s in targets[t] if s.issubset(gr)])
                if len(b)>7:
                    matches[frozenset(gr)] = b

    m: List[MatchPhasePolyType[VT]] = []
    taken: Set[VT] = set()
    for groupp, gad in sorted(matches.items(), key=lambda x: len(x[1]), reverse=True):
        if taken.intersection(groupp): continue
        m.append((list(groupp), {s:(gadgets[s] if len(s)>1 else list(s)[0]) for s in gad}))
        taken.update(groupp)

    return m

def apply_gadget_phasepoly(g: BaseGraph[VT,ET], matches: List[MatchPhasePolyType[VT]]) -> None:
    """Uses the output of :func:`match_gadgets_phasepoly` to apply a rewrite based
    on rule R_13 of the paper *A Finite Presentation of CNOT-Dihedral Operators*."""
    rs = g.rows()
    phases = g.phases()
    for group, gadgets in matches:
        for i in range(4):
            v1 = group[i]
            g.add_to_phase(v1, Fraction(5,4))

            for j in range(i+1,4):
                v2 = group[j]
                f = frozenset({v1,v2})
                if f in gadgets:
                    n,v = gadgets[f] # type: ignore # complex typing situation
                    phase = phases[v]
                    if phases[n]:
                        phase = -phase
                        g.set_phase(n,0)
                else:
                    n = g.add_vertex(VertexType.Z,-1, rs[v2]+0.5)
                    v = g.add_vertex(VertexType.Z,-2, rs[v2]+0.5)
                    phase = 0
                    g.add_edges([(n,v),(v1,n),(v2,n)],EdgeType.HADAMARD)
                g.set_phase(v, phase + Fraction(3,4))

                for k in range(j+1,4):
                    v3 = group[k]
                    f = frozenset({v1,v2,v3})
                    if f in gadgets:
                        n,v = gadgets[f] # type: ignore
                        phase = phases[v]
                        if phases[n]:
                            phase = -phase
                            g.set_phase(n,0)
                    else:
                        n = g.add_vertex(VertexType.Z,-1, rs[v3]+0.5)
                        v = g.add_vertex(VertexType.Z,-2, rs[v3]+0.5)
                        phase = 0
                        g.add_edges([(n,v),(v1,n),(v2,n),(v3,n)],EdgeType.HADAMARD)
                    g.set_phase(v, phase + Fraction(1,4))
        f = frozenset(group)
        if f in gadgets:
            n,v = gadgets[f] # type: ignore
            phase = phases[v]
            if phases[n]:
                phase = -phase
                g.set_phase(n,0)
        else:
            n = g.add_vertex(VertexType.Z, -1, rs[group[0]]+0.5)
            v = g.add_vertex(VertexType.Z, -2, rs[group[0]]+0.5)
            phase = 0
            g.add_edges([(n,v)]+[(n,w) for w in group],EdgeType.HADAMARD)
        g.set_phase(v, phase + Fraction(7,4))



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys

from ..circuit import Circuit
from ..simplify import id_simp
from .. import tikz

def to_tikz(source, target):
    circ = Circuit.load(source)
    print("Converting circuit with {:d} gates to TikZ".format(len(circ.gates)))
    g = circ.to_graph()
    id_simp(g,quiet=True)
    tikz_output = tikz.to_tikz(g)
    print("Output file: ", os.path.abspath(target))
    f = open(target, 'w')
    f.write(tikz_output)
    f.close()

helpstring = """usage: pyzx tikz source [dest]

Script for converting circuits into tikz files.

positional arguments:
   source       File containing circuit
   dest         Desired output location for TikZ file

The default value for dest is to put a .tikz file of the same name in the folder of source.
"""

def main(args):
    if not args:
        print(helpstring)
    elif len(args) == 1:
        source = args[0]
        if not os.path.exists(source):
            print("File '{}' does not exist".format(source))
        else:
            basename = os.path.splitext(source)[0]
            target = basename+".tikz"
            to_tikz(source, target)
    else:
        source = args[0]
        target = args[1]
        to_tikz(source, target)




# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Nothing in this file will make sense if you haven't read Section IV of
# https://journals.aps.org/prx/pdf/10.1103/PhysRevX.6.021043
# In particular the text below equation (10) and equation (11) itself.

from collections import defaultdict
import random
import math
sq2 = math.sqrt(2)
omega = (1+1j)/sq2
from fractions import Fraction
from typing import List, Optional, Dict, Tuple, Any

import numpy as np

from .utils import EdgeType, FractionLike, VertexType, toggle_vertex, toggle_edge, ave_pos
from . import simplify
from .circuit import Circuit
from .graph.base import BaseGraph,VT,ET
from .graph.scalar import DyadicNumber
from .symbolic import Poly

MAGIC_GLOBAL = DyadicNumber(k=2, a=-7, b=0, c=7, d=-10)
MAGIC_B60 = DyadicNumber(k=-2, a=-4, b=3, c=0, d=3)
MAGIC_B66 = DyadicNumber(k=-2, a=24, b=-17, c=0, d=-17)
MAGIC_E6 = DyadicNumber(k=0, a=10, b=-7, c=0, d=-7)
MAGIC_O6 = DyadicNumber(k=-1, a=-7, b=5, c=0, d=5)
MAGIC_K6 = DyadicNumber(k=0, a=7, b=-5, c=0, d=-5)
MAGIC_PHI = DyadicNumber(k=0, a=10, b=-7, c=0, d=-7)

class SumGraph(object):
    """Container class for a sum of ZX-diagrams"""
    graphs: List[BaseGraph]
    def __init__(self, graphs:Optional[List[BaseGraph]]=None) -> None:
        if graphs is not None:
            self.graphs = graphs
        else:
            self.graphs = []
            
    def to_tensor(self) -> np.ndarray:
        if not self.graphs: return np.zeros((1,1))
        t = self.graphs[0].to_tensor(True)
        for i in range(len(self.graphs)-1):
            t = t + self.graphs[i+1].to_tensor(True)
        return t

    def to_matrix(self) -> np.ndarray:
        if not self.graphs: return np.zeros((1,1))
        t = self.graphs[0].to_matrix(True)
        for i in range(len(self.graphs)-1):
            t = t + self.graphs[i+1].to_matrix(True)
        return t

    def full_reduce(self, quiet:bool=True) -> None:
        terms = []
        for i, g in enumerate(self.graphs):
            if not quiet:
                print("Graph {:d}:".format(i))
            simplify.full_reduce(g, quiet=quiet)
            if not g.scalar.is_zero: terms.append(g)
            elif not quiet: print("Graph {:d} is zero".format(i))
        self.graphs = terms

    def reduce_scalar(self, quiet:bool=True) -> None:
        terms = []
        for i, g in enumerate(self.graphs):
            if not quiet:
                print("Graph {:d}:".format(i))
            simplify.reduce_scalar(g, quiet=quiet)
            if not g.scalar.is_zero: terms.append(g)
            elif not quiet: print("Graph {:d} is zero".format(i))
        self.graphs = terms

    def inner_product_with_random_state(self) -> complex:
        """All the graphs should be Clifford states with same amount of outputs.
        We compose them with a random equatorial Clifford effect,
        and we calculate the resulting inner product. 
        Used in the norm estimation algorithm of
        https://arxiv.org/pdf/1808.00128.pdf"""
        terms = self.graphs
        if len(terms) == 0: return 0.0
        q = terms[0].num_outputs()
        # We want to compose g with a random equatorial Clifford effect.
        # In the ZX-calculus this consists of random k*pi/2 phases
        # with connections being a Erdos-Renyi random graph with probability 1/2.
        # We make these spiders by replacing the outputs of the graphs.
        # Generate the data for a random equatorial Clifford effect
        phases = [Fraction(random.randint(0,3),2) for _ in range(q)]
        connections = [(i1,i2) for i1 in range(q) for i2 in range(i1+1,q) if random.random() > 0.5]
        
        val: complex = 0
        for g in terms:
            g = g.copy()
            vs = g.outputs()
            for i, v in enumerate(vs):
                g.set_type(v, VertexType.Z)
                g.set_phase(v, phases[i])
            g.set_outputs(())
            g.add_edges([(vs[i1],vs[i2]) for (i1,i2) in connections],EdgeType.HADAMARD)
            g.scalar.add_power(len(connections))
            # Now that we have composed g with a right sort of effect, 
            # we need to calculate the value of the resulting inner product.
            # Since the diagram is a scalar, full_reduce() will completely annihilate it.
            simplify.full_reduce(g)
            g.remove_isolated_vertices()
            if g.num_vertices() != 0: raise Exception("Diagram wasn't fully reduced")
            val += g.scalar.to_number()
        return val

    def estimate_norm(self, epsilon:float=0.05) -> float:
        """Uses the algorithm of https://arxiv.org/pdf/1808.00128.pdf (p.22)
        to estimate the norm squared of this state."""
        count = int(4*(1/epsilon)**2)
        total = 0.0
        for _ in range(count):
            val = self.inner_product_with_random_state()
            total += abs(val)**2
        return total/count

    def post_select(self, qubits: Dict[int, str]) -> 'SumGraph':
        """Outputs a new GraphSum, where for every term we replace the post-selected
        outputs by an effects. The argument ``qubits`` should be ``{q1:e1, q2:e2,...}``
        where the ``e1,e2,`` etc. are in the set ``{'0', '1', '+', '-'}``."""
        terms = []
        for g in self.graphs:
            g = g.copy()
            outputs = g.outputs()
            for qubit, effect in qubits.items():
                v = outputs[qubit]
                g.set_type(v,VertexType.Z)
                g.scalar.add_power(-1)
                if effect in ('0', '1'): #Push a Hadamard gate out of the spider
                    e = list(g.incident_edges(v))[0]
                    et = g.edge_type(e)
                    g.set_edge_type(e,toggle_edge(et))
                if effect in ('1', '-'):
                    g.set_phase(v,1)
            g.set_outputs(tuple(v for i,v in enumerate(outputs) if i not in qubits))
            terms.append(g)
        return SumGraph(terms)

    def sample(self, 
        qubits: List[int], 
        post_selected:Optional[Dict[int,str]]=None, 
        amount:int=10, 
        epsilon:float=0.05, 
        quiet:bool=True) -> List[List[Tuple[int,int]]]:
        """Implements the weak simulation algorithm of https://arxiv.org/pdf/1808.00128.pdf.
        ``qubits`` should be a list of qubit numbers from which measurement outcomes in the 
        computational basis are to be sampled. ``post_selected`` should be in the format of
        :method:`post_select`. ``amount`` dictates the amount of samples to be taken.
        ``epsilon`` is the error used in the norm estimation.

        Example: ``gsum.sample([1,2,3], {5:'+', 6:'0'}, 20)``
        """
        if post_selected is not None: gsum = self.post_select(post_selected)
        else: gsum = self
        gsum.full_reduce()

        qubit_map = {q:q for q in qubits}
        if post_selected is not None:
            for q in qubits: #If we post-select, this changes which qubit points to which output
                qubit_map[q] -= sum(1 for v in post_selected if v<q)
        norm = gsum.estimate_norm(epsilon)
        if not quiet: print("Estimated original norm:", norm)
        if norm < 0.01 and not quiet: 
            print("Norm very close to zero. Possibly post-selected to zero probability event?")
        probs : Dict[str,float] = {}
        outputs = []
        for i in range(amount):
            if not quiet: print("Sample", i)
            prevprob = 1.0
            path = '' # which qubits have which outcomes in this 'run'
            output = []
            current = gsum
            qubit_map2 = {q:qubit_map[q] for q in qubits}
            for q in qubits:
                path += str(q)
                temp = None
                if path not in probs:
                    temp = current.post_select({qubit_map2[q]:'0'})
                    temp.full_reduce()
                    norm2 = temp.estimate_norm(epsilon)
                    probs[path] = (norm2/norm)/prevprob
                    if not quiet: print("New prob:", "path", path, "norm", norm2, "prob", probs[path])
                prob = probs[path]
                val = int(random.random() > prob) # Actually make a sample
                output.append((q,val))
                path += str(val)

                if val == 0:
                    if temp is not None: current = temp
                    else: current = current.post_select({qubit_map2[q]:'0'})
                    prevprob *= prob
                else:
                    current = current.post_select({qubit_map2[q]:'1'})
                    prevprob *= 1-prob
                for h in qubits:
                    if h>q: qubit_map2[h] -= 1
            if not quiet: print(output)

            outputs.append(output)
        return outputs
                

                

def calculate_path_sum(g: BaseGraph[VT,ET]) -> complex:
    """Input should be a fully reduced scalar graph-like Clifford+T ZX-diagram. 
    Calculates the scalar it represents."""
    if g.num_vertices() < 2: return g.to_tensor().flatten()[0]
    phases = g.phases()
    prefactor = 0
    variable_dict : Dict[VT,int] = dict()
    variables: List[int] = [] # Contains the phases of each of the variables
    czs = []
    xors = dict()
    for v in g.vertices():
        if v in variable_dict: continue
        if not phases[v]: continue #It is the axle of a phase gadget, ignore it
        if g.vertex_degree(v) == 1: # Probably phase gadget
            w = list(g.neighbors(v))[0]
            if not phases[w]: # It is indeed a phase gadget
                targets = set()
                for t in g.neighbors(w):
                    if t == v: continue
                    if t not in variable_dict:
                        variable_dict[t] = len(variables)
                        phase_t = phases[t]
                        if isinstance(phase_t, Poly):
                            raise NotImplementedError("Symbolic phases not supported")
                        variables.append(int(float(phase_t*4)))
                    targets.add(variable_dict[t])
                prefactor += len(targets)-1
                phase_v = phases[v]
                if isinstance(phase_v, Poly):
                    raise NotImplementedError("Symbolic phases not supported")
                xors[frozenset(targets)] = int(float(phase_v*4))
                continue
        variable_dict[v] = len(variables)
        phase_v = phases[v]
        if isinstance(phase_v, Poly):
            raise NotImplementedError("Symbolic phases not supported")
        variables.append(int(float(phase_v*4)))
    verts = sorted(list(variable_dict.keys()), key=lambda x: variable_dict[x])
    n = len(verts)
    for i in range(n):
        v = verts[i]
        for j in range(i+1,n):
            w = verts[j]
            if g.connected(v,w):
                czs.append((i,j))
                prefactor += 1

    c = Circuit(n)
    for i in range(n):
        c.add_gate("ZPhase",i,phase=Fraction(variables[i],4))
    for tgts, phase in xors.items():
        c.add_gate("ParityPhase", Fraction(phase,4), *list(tgts))
    for i,j in czs:
        c.add_gate("CZ", i,j)
    g2 = c.to_graph()
    g2.apply_state("+"*n)
    g2.apply_effect("+"*n)
    g2.scalar.add_power(2*n)
    val = g2.to_tensor().flatten()[0]
    return g.scalar.to_number()*val*(sq2**(-prefactor))
    # variables = np.array(variables)
    # xors = [(np.array([int(k in xor) for k in range(n)]),phase) for xor,phase in xors.items()]
    # print(n,len(czs),len(xors))
    
    # results = [0]*8
    # for bitstring in itertools.product([0,1],repeat=n):
    #     val = np.dot(variables,bitstring)
    #     val += sum(phase*(np.dot(bitstring,xor)%2) for xor,phase in xors)
    #     val += 4*sum(1 for i,j in czs if bitstring[i] and bitstring[j])
    #     try: results[val % 8] += 1
    #     except: print(val)
    # print(results)
    # r = results
    # return g.scalar.to_number()*sq2**(-prefactor)*(r[0]-r[4]+omega*(r[1]-r[5]) +1j*(r[2]-r[6]) + 1j*omega*(r[3]-r[7]))

def find_stabilizer_decomp(g: BaseGraph[VT,ET]) -> List[BaseGraph[VT,ET]]:
    if simplify.tcount(g) == 0: return [g]
    gsum = replace_magic_states(g, False)
    gsum.reduce_scalar()
    output = []
    for h in gsum.graphs:
        if h.scalar.is_zero: continue
        output.extend(find_stabilizer_decomp(h))
    return output

def max_terms_needed(g: BaseGraph[VT,ET]) -> int:
    """Returns the maximum amount of stabilizer terms that g could be split in
    by :func:``find_stabilizer_decomp``."""
    v = simplify.tcount(g)
    count = 7**(v//6)
    v -= 6*(v//6)
    if v == 5: return count*8
    if v in (4,3): return count*4
    if v in (2,1): return count*2
    return count


def replace_magic_states(g: BaseGraph[VT,ET], pick_random:Any=False) -> SumGraph:
    """This function takes in a ZX-diagram in graph-like form 
    (all spiders fused, only Z spiders, only H-edges between spiders),
    and splits it into a sum over smaller diagrams by using the magic
    state decomposition of Bravyi, Smith, and Smolin (2016), PRX 6, 021043.
    """
    g = g.copy() # We copy here, so that the vertex labels we get will be the same ones if we copy the graph again
    phases = g.phases()

    # First we find 6 T-like spiders
    boundary = []
    internal = []
    gadgets = []
    ranking: Dict[VT, int] = dict()
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        if not phases[v] or phases[v].denominator != 4: continue

        ### begin AK changes ....
        deg = g.vertex_degree(v)
        if g.vertex_degree(v) == 1:
            w = list(g.neighbors(v))[0]
            if g.type(w) == VertexType.Z:
                gadgets.append(v)
                deg = g.vertex_degree(w)-1

        if any(w in inputs or w in outputs for w in g.neighbors(v)):
            boundary.append(v)
        else:
            internal.append(v)
        ranking[v] = deg
        ### ... end AK changes

    if len(ranking) >= 6: num_replace = 6
    elif len(ranking) >= 2: num_replace = 2
    elif len(ranking) == 1: num_replace = 1
    else: raise Exception("No magic states to replace")

    if not pick_random:
        candidates = sorted(ranking.keys(), key=lambda v: ranking[v], reverse=True)[:num_replace]
    else:
        if not isinstance(pick_random,bool):
            random.seed(pick_random)
        candidates = random.sample(list(ranking.keys()),num_replace)

    graphs = []
    if num_replace == 6:
        replace_functions = [replace_B60, replace_B66, replace_E6, replace_O6, replace_K6, replace_phi1, replace_phi2]
    elif num_replace == 2:
        replace_functions = [replace_2_S, replace_2_N]
    else:
        replace_functions = [replace_1_0, replace_1_1]

    for func in replace_functions:
        h = func(g.copy(), candidates)
        if num_replace == 6: h.scalar.add_float(MAGIC_GLOBAL)
        graphs.append(h)

    return SumGraph(graphs)

def replace_B60(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_B60)
    g.scalar.add_power(-6)
    for v in verts:
        g.add_to_phase(v,Fraction(-1,4))
    return g

def replace_B66(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_B66)
    g.scalar.add_power(-6)
    g.scalar.add_phase(Fraction(1))
    for v in verts:
        g.add_to_phase(v,Fraction(-1,4))
        g.add_to_phase(v,Fraction(1))
    return g

def replace_E6(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_E6)
    g.scalar.add_power(4)
    g.scalar.add_phase(Fraction(1,2))
    av = 0.0
    for v in verts:
        g.add_to_phase(v,Fraction(-1,4))
        g.add_to_phase(v, Fraction(1,2))
        av += g.row(v)
    w = g.add_vertex(VertexType.Z,-1,av/6,Fraction(1))
    g.add_edges([(v,w) for v in verts],EdgeType.HADAMARD)
    return g

def replace_O6(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_O6)
    g.scalar.add_power(4)
    g.scalar.add_phase(Fraction(1,2))
    av = 0.0
    for v in verts:
        g.add_to_phase(v,Fraction(-1,4))
        g.add_to_phase(v, Fraction(1,2))
        av += g.row(v)
    w = g.add_vertex(VertexType.Z,-1,av/6,Fraction(0))
    g.add_edges([(v,w) for v in verts],EdgeType.HADAMARD)
    return g

def replace_K6(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_K6)
    g.scalar.add_power(5)
    g.scalar.add_phase(Fraction(1,4))
    av = 0.0
    for v in verts:
        g.add_to_phase(v,Fraction(-1,4))
        av += g.row(v)
    w = g.add_vertex(VertexType.Z,-1,av/6,Fraction(3,2))
    g.add_edges([(v,w) for v in verts],EdgeType.SIMPLE)
    return g

def replace_phi1(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_float(MAGIC_PHI)
    g.scalar.add_power(9)
    g.scalar.add_phase(Fraction(3,2))
    w6 = g.add_vertex(VertexType.Z,-1, g.row(verts[5])+0.5, Fraction(1))
    g.add_to_phase(verts[5],Fraction(-1,4))
    g.add_edge((verts[5],w6))
    ws = []
    for v in verts[:-1]:
        g.add_to_phase(v,Fraction(-1,4))
        w = g.add_vertex(VertexType.Z,-1, g.row(v)+0.5)
        g.add_edges([(w6,w),(v,w)],EdgeType.HADAMARD)
        ws.append(w)
    w1,w2,w3,w4,w5 = ws
    g.add_edges([(w1,w3),(w1,w4),(w2,w4),(w2,w5),(w3,w5)],EdgeType.HADAMARD)
    return g

def replace_phi2(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    v1,v2,v3,v4,v5,v6 = verts
    verts = [v1,v2,v4,v5,v6,v3]
    return replace_phi1(g,verts)


def replace_2_S(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, Fraction(1,2))
    g.add_edges([(verts[0],w),(verts[1],w)],EdgeType.SIMPLE)
    for v in verts: g.add_to_phase(v,Fraction(-1,4))
    return g


def replace_2_N(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_phase(Fraction(1,4))
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, Fraction(1,1))
    g.add_edges([(verts[0],w),(verts[1],w)],EdgeType.HADAMARD)
    for v in verts: g.add_to_phase(v,Fraction(-1,4))
    return g

def replace_1_0(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_power(-1)
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, 0)
    g.add_edge((verts[0],w),EdgeType.HADAMARD)
    for v in verts: g.add_to_phase(v,Fraction(-1,4))
    return g

def replace_1_1(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_phase(Fraction(1,4))
    g.scalar.add_power(-1)
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, Fraction(1,1))
    g.add_edge((verts[0],w),EdgeType.HADAMARD) 
    for v in verts: g.add_to_phase(v,Fraction(-1,4))
    return g


def replace_1_0_arbitrary_rot(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    g.scalar.add_power(-1)
    phase = g.phase(verts[0])
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, 0)
    g.add_edge((verts[0],w),EdgeType.HADAMARD)
    for v in verts: g.add_to_phase(v,-phase)
    return g

def replace_1_1_arbitrary_rot(g: BaseGraph[VT,ET], verts: List[VT]) -> BaseGraph[VT,ET]:
    phase = g.phase(verts[0])
    g.scalar.add_phase(phase)
    g.scalar.add_power(-1)
    w = g.add_vertex(VertexType.Z,g.qubit(verts[0])-0.5, g.row(verts[0])-0.5, Fraction(1,1))
    g.add_edge((verts[0],w),EdgeType.HADAMARD) 
    for v in verts: g.add_to_phase(v,-phase)
    return g

def replace_u3_phases(g: BaseGraph[VT,ET]) -> SumGraph:
    for v in g.vertices():
        if g.phase(v).denominator not in (1, 2, 4):
            g1 = replace_1_1_arbitrary_rot(g.copy(), [v])
            g2 = replace_1_0_arbitrary_rot(g.copy(), [v])
            return SumGraph([g1, g2])
    return SumGraph([g])

def cut_vertex(g,v):
    """Applies the ``cutting'' decomposition to a vertex, as used in, for example: https://arxiv.org/pdf/2403.10964."""
    g  = g.clone()
    g0 = g.clone()
    g1 = g.clone()
    g0.remove_vertex(v)
    g1.remove_vertex(v)

    n = len(g.neighbors(v))
    g0.scalar.add_power(-n)
    g1.scalar.add_power(-n)
    g1.scalar.add_phase(g.phase(v)) # account for e^(i*pi*alpha) on right branch
    
    vtype = toggle_vertex(g.type(v))

    for i in g.neighbors(v):
        etype = g.edge_type((v,i)) # maintain edge type
        qubit = ave_pos(g.qubit(v),g.qubit(i),1/2)
        row   = ave_pos(g.row(v),g.row(i),1/2)
        
        newV = g0.add_vertex(vtype,qubit,row,0) # add and connect the new vertices
        g0.add_edge((newV,i),etype)
        newV = g1.add_vertex(vtype,qubit,row,1)
        g1.add_edge((newV,i),etype)

    return (g0,g1)

def cut_edge(g,e,ty=1):
    """Applies the ``cutting'' decomposition to an edge, as used in, for example: https://arxiv.org/pdf/2403.10964. The type ty decides whether to cut with Z- branches or X- branches."""
    g  = g.clone()
    g0 = g.clone()
    g1 = g.clone()
    g0.remove_edge(e)
    g1.remove_edge(e)

    etype = g.edge_type(e)

    g0.scalar.add_power(-2)
    g1.scalar.add_power(-2)

    x0,x1 = g.row(e[0]), g.row(e[1])
    y0,y1 = g.qubit(e[0]), g.qubit(e[1])

    qubit1 = ave_pos(y0,y1,1/3)
    row1   = ave_pos(x0,x1,1/3)
    qubit2 = ave_pos(y0,y1,2/3)
    row2   = ave_pos(x0,x1,2/3)
    
    v = g0.add_vertex(ty=ty,qubit=qubit1,row=row1,phase=0)
    g0.add_edge((v,e[0]),1)
    v = g0.add_vertex(ty=ty,qubit=qubit2,row=row2,phase=0)
    g0.add_edge((v,e[1]),etype)

    v = g1.add_vertex(ty=ty,qubit=qubit1,row=row1,phase=1)
    g1.add_edge((v,e[0]),1)
    v = g1.add_vertex(ty=ty,qubit=qubit2,row=row2,phase=1)
    g1.add_edge((v,e[1]),etype)
    
    return (g0,g1)

def cut_wishbone(g,v,neighs,ph):
    """Applies the ``wishbone cut'' (or ``separator cut'') decomposition to vertex v of graph g, pulling out the neighbours ``neighs'' and a phase ``ph'', as used in: [PAPER UPCOMING]."""
    g = g.clone()
    
    for i in neighs:
        if not i in g.neighbors(v):
            raise ValueError("Attempted illegal wishbone cut. Vertex " + str(i) + " is not a neighbor of target vertex " + str(v) + ".")
    
    neighs_left  = set(g.neighbors(v)).symmetric_difference(neighs)
    neighs_right = neighs
    
    phase_left  = g.phase(v) - ph
    phase_right = ph
    
    v_left  = g.add_vertex(qubit=g.qubit(v),row=g.row(v)-0.5,ty=g.type(v),phase=phase_left)
    v_right = g.add_vertex(qubit=g.qubit(v),row=g.row(v)+0.5,ty=g.type(v),phase=phase_right)
    
    for i in neighs_left:  g.add_edge((v_left,i),g.edge_type((v,i)))
    for i in neighs_right: g.add_edge((v_right,i),g.edge_type((v,i)))
    
    g.remove_vertex(v)
    
    #--
    
    gLeft  = g.clone()
    gRight = g.clone()
    
    gRight.add_to_phase(v_left,1)
    gRight.add_to_phase(v_right,1)
    
    return (gLeft,gRight)


def gen_catlike_term(g_initial: BaseGraph[VT, ET],
                     vertices: List[VT],
                     ph_base: FractionLike,
                     ph_central: FractionLike,
                     ph_appendix: FractionLike,
                     eType_base: EdgeType,
                     eType_appendix: EdgeType,
                     scal_positive: bool,
                     scal_power: int,
                     scal_phase: FractionLike,
                     pi_case: bool = False) -> BaseGraph[VT, ET]:
    """Insert a term from a cat or magic5 decomposition into a graph.

    Used for constructing graph terms with local structure of the form of the
    right-hand side terms of the cat (and magic5) decompositions seen in
    https://arxiv.org/pdf/2202.09202.

    For example, consider the magic5 decomposition in this above paper. Here,
    we exchange 5 T-spiders (verts) of our graph (gInit) for three catlike
    terms.

    The third and final such term has:
        - Phase pi/2 on each of the 5 outgoing edges,
        i.e. ph_base=Fraction(1, 2)
        - Phase 0 on the inner central spider to which the others are
        connected, i.e. ph_central=0
        - Phase pi/4 on the outstanding one-legged 'appendix' spider,
        i.e. ph_appendix=Fraction(1, 4)
        - Hadamard edges connecting each of the outgoing spiders to the inner
        central spider, i.e. eType_base=EdgeType.HADAMARD
        - Hadamard edge connecting the appendix spider to the central spider,
        i.e. eType_appendix=EdgeType.HADAMARD.
        - A negative sign on the scalar factor, i.e. scal_positive=False
        - A sqrt(2)^3 factor in its scalar, i.e. scal_power=3
        - An e^{i*pi/4} factor in its scalar, i.e. scal_phase=Fraction(1, 4)

    The `pi_case` parameter is used to handle the case where the phase of the
    Clifford spider in the state being decomposed is pi.
    """
    g = g_initial.clone()

    # Handle central phase for pi case
    if pi_case:
        ph_central += 1

    # Add the scalar factor
    g.scalar.add_power(scal_power)
    g.scalar.add_phase(scal_phase if scal_positive else scal_phase + 1)

    # Add the central and appendix vertices
    new_v1 = g.add_vertex(qubit=-1, row=-1, ty=VertexType.Z, phase=ph_central)
    new_v2 = g.add_vertex(qubit=-2, row=-1, ty=VertexType.Z, phase=ph_appendix)

    # Connect the appendix vertex to the central vertex
    g.add_edge((new_v1, new_v2), eType_appendix)

    # Connect the central vertex to the outgoing vertices
    # and give each outgoing vertex the appropriate phase
    for i, v in enumerate(vertices):
        phase_change = ph_base
        # Handle pi case. We make an arbitrary choice of the last vertex
        # as where to apply the correction.
        if pi_case and i == len(vertices) - 1:
            phase_change -= Fraction(1, 2)
            phase_change *= -1
        # Subtracting pi/4 and adding ph_base is equivalent to
        # unfusing, decomposing, then refusing.
        phase_change -= Fraction(1, 4)

        g.add_to_phase(v, phase_change)
        g.add_edge((v, new_v1), eType_base)

    return g


def check_catn(g: BaseGraph[VT, ET], vertex: VT, n: int) -> bool:
    """Ensure cat `n` decomposition is applicable to a vertex of graph `g`.

    See https://arxiv.org/pdf/2202.09202.
    """
    if n not in (3, 4, 5, 6):
        raise ValueError(f'Cat {n} decomposition not implemented. '
                         + 'Try cat3, cat4, cat5, or cat6 instead.')

    if vertex not in g.vertices():
        raise ValueError(f'Vertex {vertex} does not exist in graph')

    phase = g.phase(vertex)
    if phase % 1 != 0:
        raise ValueError(f'The cat {n} decomposition function can only be '
                         + 'applied to a phaseless or pi-phase spider, but '
                         + f'specified vertex has phase {phase}')
    neighbors = g.neighbors(vertex)
    if len(neighbors) != n:
        raise ValueError(f'The cat {n} decomposition acts on a degree {n} '
                         + 'spider, but specified vertex has degree '
                         + str(len(neighbors)))

    vertex_type = g.type(vertex)
    for neighbor in neighbors:
        neighbor_type = g.type(neighbor)
        edge_type = g.edge_type(g.edge(vertex, neighbor))
        if (
          (edge_type != EdgeType.HADAMARD
           or neighbor_type != vertex_type)
          and
          (edge_type != EdgeType.SIMPLE
           or neighbor_type != toggle_vertex(vertex_type))
          ):
            raise ValueError(
                f'The cat {n} decomposition must act on a spider with {n} '
                + 'opposite colour neighbours '
                + '(or like-coloured with Hadamard edges).')

    return True


def apply_magic5(g: BaseGraph[VT,ET], verts: List[VT]) -> SumGraph:
    """Apply the magic5 decomposition to vertices verts of graph g. See: https://arxiv.org/pdf/2202.09202."""
    assert len(verts) == 5, "The magic5 decomposition acts on 5 spiders but you specified "+str(len(verts))
    for v in verts: assert v in g.vertices(), "Vertex "+str(v)+" does not exist in graph"
    
    g_A = gen_catlike_term(g, verts, 0,              Fraction(-1,2), Fraction(-1,4), EdgeType.SIMPLE,   EdgeType.SIMPLE,   True,  2, 0)
    g_B = gen_catlike_term(g, verts, 0,              0,              Fraction(-1,4), EdgeType.HADAMARD, EdgeType.HADAMARD, True,  3, Fraction(3,4))
    g_C = gen_catlike_term(g, verts, Fraction(1,2),  0,              Fraction(1,4),  EdgeType.HADAMARD, EdgeType.HADAMARD, False, 3, Fraction(1,4))
    
    return SumGraph([g_A, g_B, g_C])


def apply_cat3(g: BaseGraph[VT, ET], vertex: VT) -> SumGraph:
    """Apply the cat3 decomposition to a vertex of graph `g`.

    See https://arxiv.org/pdf/2202.09202."""
    # Check the decomposition is applicable
    check_catn(g, vertex, 3)
    # Generate the terms of the decomposition
    neighbors = list(g.neighbors(vertex))
    pi_case = g.phase(vertex) == 1
    g_A = gen_catlike_term(g, neighbors,
                           0, Fraction(-1, 2), 0,
                           EdgeType.SIMPLE, EdgeType.HADAMARD,
                           True, -1, Fraction(-1, 4), pi_case=pi_case)
    g_B = gen_catlike_term(g, neighbors,
                           0, 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           True,  0, Fraction(1, 2), pi_case=pi_case)
    # Remove the decomposed vertices from the terms
    g_A.remove_vertex(vertex)
    g_B.remove_vertex(vertex)

    return SumGraph([g_A, g_B])


def apply_cat4(g: BaseGraph[VT, ET], vertex: VT) -> SumGraph:
    """Apply the cat4 decomposition to a vertex of graph `g`.

    See https://arxiv.org/pdf/2202.09202."""
    # Check the decomposition is applicable
    check_catn(g, vertex, 4)
    # Generate the terms of the decomposition
    neighbors = list(g.neighbors(vertex))
    pi_case = g.phase(vertex) == 1
    g_A = gen_catlike_term(g, neighbors,
                           0, Fraction(-1, 2), 0,
                           EdgeType.SIMPLE, EdgeType.SIMPLE,
                           True, -1, Fraction(-1, 4), pi_case=pi_case)
    g_B = gen_catlike_term(g, neighbors,
                           0, 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           True,  0, Fraction(1, 2), pi_case=pi_case)
    # Remove the decomposed vertices from the terms
    g_A.remove_vertex(vertex)
    g_B.remove_vertex(vertex)

    return SumGraph([g_A, g_B])


def apply_cat5(g: BaseGraph[VT, ET], vertex: VT) -> SumGraph:
    """Apply the cat5 decomposition to a vertex of graph `g`.

    See https://arxiv.org/pdf/2202.09202."""
    # Check the decomposition is applicable
    check_catn(g, vertex, 5)
    # Generate the terms of the decomposition
    neighbors = list(g.neighbors(vertex))
    pi_case = g.phase(vertex) == 1
    g_A = gen_catlike_term(g, neighbors,
                           0, Fraction(-1, 2), 0,
                           EdgeType.SIMPLE, EdgeType.HADAMARD,
                           True,  -2, 0, pi_case=pi_case)
    g_B = gen_catlike_term(g, neighbors,
                           0, 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           True,  -1, Fraction(3, 4), pi_case=pi_case)
    g_C = gen_catlike_term(g, neighbors,
                           Fraction(1, 2), 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           False, -1, Fraction(1, 4), pi_case=pi_case)
    # Remove the decomposed vertices from the terms
    g_A.remove_vertex(vertex)
    g_B.remove_vertex(vertex)
    g_C.remove_vertex(vertex)

    return SumGraph([g_A, g_B, g_C])


def apply_cat6(g: BaseGraph[VT, ET], vertex: VT) -> SumGraph:
    """Apply the cat6 decomposition to a vertex of graph `g`.

    See https://arxiv.org/pdf/2202.09202."""
    # Check the decomposition is applicable
    check_catn(g, vertex, 6)
    # Generate the terms of the decomposition
    verts = list(g.neighbors(vertex))
    pi_case = g.phase(vertex) == 1
    g_A = gen_catlike_term(g, verts,
                           0, Fraction(-1, 2), 0,
                           EdgeType.SIMPLE, EdgeType.SIMPLE,
                           True,  -2, 0, pi_case=pi_case)
    g_B = gen_catlike_term(g, verts,
                           0, 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           True,  -1, Fraction(3, 4), pi_case=pi_case)
    g_C = gen_catlike_term(g, verts,
                           Fraction(1, 2), 0, 0,
                           EdgeType.HADAMARD, EdgeType.SIMPLE,
                           False, -1, Fraction(1, 4), pi_case=pi_case)
    # Remove the decomposed vertices from the terms
    g_A.remove_vertex(vertex)
    g_B.remove_vertex(vertex)
    g_C.remove_vertex(vertex)

    return SumGraph([g_A, g_B, g_C])

def replace_1_0_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    g.scalar.add_power(-1)
    phase = g.phase(verts[0])
    w = g.add_vertex(VertexType.Z, g.qubit(verts[0]) - 0.5, g.row(verts[0]) - 0.5, 0)
    g.add_edge(g.edge(verts[0], w), EdgeType.HADAMARD)
    for v in verts:
        g.add_to_phase(v, -phase)
    return g


def replace_1_1_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    phase = g.phase(verts[0])
    g.scalar.approximate_floatfactor *= np.exp(1j * phase * np.pi)
    g.scalar.add_power(-1)
    w = g.add_vertex(
        VertexType.Z, g.qubit(verts[0]) - 0.5, g.row(verts[0]) - 0.5, Fraction(1, 1)
    )
    g.add_edge(g.edge(verts[0], w), EdgeType.HADAMARD)
    for v in verts:
        g.add_to_phase(v, -phase)
    return g

def replace_2_bell_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    v0, v1 = verts[:2]
    phase0 = g.phase(v0) % Fraction(1, 2)
    g.add_to_phase(v0, -phase0)
    g.add_to_phase(v1, phase0)
    g.add_edge(g.edge(v0, v1), EdgeType.SIMPLE)
    return g


def replace_2_10_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    v0, v1 = verts[:2]
    phase0 = g.phase(v0) % Fraction(1, 2)
    g.add_to_phase(v0, -phase0)
    g.add_to_phase(v1, phase0)
    v0_ = g.add_vertex(
        VertexType.Z, g.qubit(v0) - 0.5, g.row(v0) - 0.5, phase=Fraction(1, 1)
    )
    v1_ = g.add_vertex(
        VertexType.Z, g.qubit(v1) - 0.5, g.row(v1) - 0.5, phase=Fraction(0, 1)
    )
    g.add_edge(g.edge(v0_, v0), EdgeType.HADAMARD)
    g.add_edge(g.edge(v1_, v1), EdgeType.HADAMARD)
    g.scalar.add_phase(phase0)
    g.scalar.add_power(-2)
    return g


def replace_2_01_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    v0, v1 = verts[:2]
    phase0 = g.phase(v0) % Fraction(1, 2)
    g.add_to_phase(v0, -phase0)
    g.add_to_phase(v1, phase0)
    v0_ = g.add_vertex(
        VertexType.Z, g.qubit(v0) - 0.5, g.row(v0) - 0.5, phase=Fraction(0, 1)
    )
    v1_ = g.add_vertex(
        VertexType.Z, g.qubit(v1) - 0.5, g.row(v1) - 0.5, phase=Fraction(1, 1)
    )
    g.add_edge(g.edge(v0_, v0), EdgeType.HADAMARD)
    g.add_edge(g.edge(v1_, v1), EdgeType.HADAMARD)
    g.scalar.add_phase(-phase0)
    g.scalar.add_power(-2)
    return g


def replace_5_0_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    theta = g.phase(verts[0]) % Fraction(1, 2)
    for v in verts:
        g.add_to_phase(v, -theta)
        v_ = g.add_vertex(VertexType.Z, g.qubit(v) - 0.5, g.row(v) - 0.5)
        g.add_edge(g.edge(v_, v), EdgeType.HADAMARD)

    g.scalar.approximate_floatfactor *= 1 - np.exp(1j * 4 * theta * np.pi)
    g.scalar.add_power(-5)
    return g


def replace_5_1_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    theta = g.phase(verts[0]) % Fraction(1, 2)
    for v in verts:
        g.add_to_phase(v, -theta)
        v_ = g.add_vertex(
            VertexType.Z, g.qubit(v) - 0.5, g.row(v) - 0.5, phase=Fraction(1, 1)
        )
        g.add_edge(g.edge(v_, v), EdgeType.HADAMARD)

    g.scalar.approximate_floatfactor *= (
        np.exp(1j * 6 * theta * np.pi) - np.exp(1j * 2 * theta * np.pi)
    ) * np.exp(-1j * theta * np.pi)
    g.scalar.add_power(-5)
    return g


def replace_5_e_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    theta = g.phase(verts[0]) % Fraction(1, 2)
    v1 = g.add_vertex(VertexType.Z, g.qubit(verts[0]) - 0.5, g.row(verts[0]) - 0.5)
    v2 = g.add_vertex(
        VertexType.Z, g.qubit(verts[1]) - 0.5, g.row(verts[1]) - 0.5, phase=-theta
    )
    g.add_edge(g.edge(v2, v1), EdgeType.HADAMARD)
    for v in verts:
        g.add_to_phase(v, -theta)
        g.add_edge(g.edge(v, v1), EdgeType.HADAMARD)
    g.scalar.approximate_floatfactor *= np.exp(1j * 3 * theta * np.pi) * np.cos(
        theta * np.pi
    )
    g.scalar.add_power(4)
    return g


def replace_5_k_arbitrary_rotation(
    g: BaseGraph[VT, ET], verts: List[VT]
) -> BaseGraph[VT, ET]:
    theta = g.phase(verts[0]) % Fraction(1, 2)
    v1 = g.add_vertex(VertexType.Z, g.qubit(verts[0]) - 0.5, g.row(verts[0]) - 0.5)
    v2 = g.add_vertex(
        VertexType.Z,
        g.qubit(verts[1]) - 0.5,
        g.row(verts[1]) - 0.5,
        phase=Fraction(1, 2) - theta,
    )
    g.add_edge(g.edge(v2, v1), EdgeType.HADAMARD)
    for v in verts:
        g.add_to_phase(v, Fraction(1, 2) - theta)
        g.add_edge(g.edge(v, v1), EdgeType.HADAMARD)
    g.scalar.approximate_floatfactor *= (
        np.exp(1j * 3 * theta * np.pi) * 1j * np.sin(theta * np.pi)
    )
    g.scalar.add_power(4)
    return g


def replace_u3_states(g: BaseGraph[VT, ET], pick_random: Any = False) -> SumGraph:
    """This function takes in a ZX-diagram in graph-like form
    (all spiders fused, only Z spiders, only H-edges between spiders),
    and splits it into a sum over smaller diagrams by using the magic
    state decomposition of Bravyi, Smith, and Smolin (2016), PRX 6, 021043.
    """
    # copy so that the vertex labels we get will be the same ones if we copy the graph again
    g = g.copy()

    # First we find vertices with arbitary phase rotations, sorted by phase
    phases: dict[Fraction, list[VT]] = defaultdict(list)
    for v in g.vertices():
        phase = g.phase(v)
        if not phase or phase.denominator in (1, 2, 4):
            continue
        phases[phase % Fraction(1, 2)].append(v)

    if len(phases) == 0:
        raise Exception("No magic states to replace")

    phase_with_most_vertices = max(phases.keys(), key=lambda x: len(phases[x]))
    vs = phases[phase_with_most_vertices]

    conjugate_phase_pair: list | None = None
    for p in phases:
        p_conj = (-p) % Fraction(1, 2)
        if p_conj in phases:
            conjugate_phase_pair = [phases[p][0], phases[p_conj][0]]
            break

    if len(vs) >= 5:
        candidates = vs[:5]
        replace_functions = [
            replace_5_0_arbitrary_rotation,
            replace_5_1_arbitrary_rotation,
            replace_5_e_arbitrary_rotation,
            replace_5_k_arbitrary_rotation,
        ]
    elif conjugate_phase_pair:
        candidates = conjugate_phase_pair
        replace_functions = [
            replace_2_bell_arbitrary_rotation,
            replace_2_10_arbitrary_rotation,
            replace_2_01_arbitrary_rotation,
        ]
    else:
        candidates = vs[0:1]
        replace_functions = [
            replace_1_0_arbitrary_rotation,
            replace_1_1_arbitrary_rotation,
        ]

    graphs = []
    for func in replace_functions:
        h = func(g.copy(), candidates)

        # import tsim.external.pyzx as zx
        # zx.draw(h, labels=True)
        graphs.append(h)

    return SumGraph(graphs)


# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from . import Circuit
from .gates import *

def parse_qc(data: str) -> Circuit:
    """Produces a :class:`Circuit` based on a .qc description of a circuit.
    If a Tofolli gate with more than 2 controls is encountered, ancilla qubits are added.
    Currently up to 5 controls are supported."""
    preamble = data[:data.find("BEGIN")].strip().splitlines()
    labels: Dict[str,int] = {}
    for l in preamble:
        if l.startswith('#'): continue
        if l.startswith('.'):
            for v in l[2:].replace(',',' ').strip().split():
                s = v.strip()
                if s not in labels: labels[s] = len(labels)
        else:
            raise TypeError("Unknown Expression: " + l)

    ancillas: Dict[int,int] = {}
    gates: List[Gate] = []
    qcount = len(labels)

    for l in data[data.find("BEGIN")+6:data.find("END")].splitlines():
        if l.startswith('#'): continue
        l = l.strip()
        if not l: continue
        try: gname, targetstr = l.split(' ',1)
        except ValueError:
            raise ValueError("Couldn't parse line {}".format(l))
        gname = gname.strip().lower()
        targets = [labels[v.strip()] for v in targetstr.replace(',',' ').strip().split(' ') if v.strip()]
        if len(targets) == 1:
            t = targets[0]
            if gname in ('tof', 't1', 'not', 'x'): gates.append(NOT(t))
            elif gname == 'z': gates.append(Z(t))
            elif gname in ('s', 'p'): gates.append(S(t))
            elif gname in ('s*', 'p*'): gates.append(S(t,adjoint=True))
            elif gname == 't': gates.append(T(t))
            elif gname == 't*': gates.append(T(t,adjoint=True))
            elif gname == 'h': gates.append(HAD(t))
            else:
                raise TypeError("Unknown gate with single target: " + l)
        elif len(targets) == 2:
            c,t = targets
            if gname in ('cnot', 'tof', 't2'):
                gates.append(CNOT(c,t))
            elif gname in ('cz', 'z'):
                gates.append(CZ(c,t))
            elif gname in ('swap', ):
                gates.append(SWAP(c,t))
            else:
                raise TypeError("Unknown gate with control: " + l)
        elif len(targets) == 3:
            c1,c2,t = targets
            if gname in ('t3', 'tof'):
                gates.append(Tofolli(c1,c2,t))
            elif gname in ('ccz', 'z'):
                gates.append(CCZ(c1,c2,t))
            else:
                raise TypeError("Unknown gate with control: " + l)
        else:
            if gname not in ('t4', 't5', 't6', 't7', 'tof'):
                raise TypeError("Unknown gate with multiple controls: " + l)
            *ctrls, t = targets
            if len(ctrls) > 6: raise TypeError("No more than 5 ctrls supported")
            while len(ancillas) < len(ctrls) - 2:
                ancillas[len(ancillas)] = qcount
                qcount += 1
            gates.append(Tofolli(ctrls[0],ctrls[1],ancillas[0]))
            if len(ctrls) == 3:
                gates.append(Tofolli(ctrls[2],ancillas[0],t))
            else:
                gates.append(Tofolli(ctrls[2],ctrls[3],ancillas[1]))
                if len(ctrls) == 4:
                    gates.append(Tofolli(ancillas[0],ancillas[1],t))
                elif len(ctrls) == 5:
                    gates.append(Tofolli(ancillas[0],ancillas[1],ancillas[2]))
                    gates.append(Tofolli(ctrls[4],ancillas[2],t))
                    gates.append(Tofolli(ancillas[0],ancillas[1],ancillas[2]))
                else: # len(ctrls) == 6
                    gates.append(Tofolli(ctrls[4],ctrls[5],ancillas[2]))
                    gates.append(Tofolli(ancillas[0],ancillas[1],ancillas[3]))
                    gates.append(Tofolli(ancillas[2],ancillas[3],t))
                    gates.append(Tofolli(ancillas[0],ancillas[1],ancillas[3]))
                    gates.append(Tofolli(ctrls[4],ctrls[5],ancillas[2]))
                gates.append(Tofolli(ctrls[2],ctrls[3],ancillas[1]))
            gates.append(Tofolli(ctrls[0],ctrls[1],ancillas[0]))

    circ = Circuit(qcount)
    circ.gates = gates
    return circ




# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2025 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyzx_param.graph.multigraph import Edge
from .utils import EdgeType, VertexType
from .graph.base import BaseGraph, VT, ET
from itertools import combinations
from fractions import Fraction

def check_fourier(g: BaseGraph[VT,ET], v: VT) -> bool:
    """Returns True if the given node is an H-box adjacent to Z spiders"""
    ty = g.types()
    if (ty[v] == VertexType.H_BOX and
        all(ty[w] == VertexType.Z for w in g.neighbors(v)) and 
        all(g.edge_type(e) == EdgeType.SIMPLE for e in g.incident_edges(v))):
        return True
    else:
        return False

def fourier(g: BaseGraph[VT, ET], v: VT, graph_like: bool=False) -> bool:
    """Applies the graphical fourier transform, replacing the given H-box of arity k with
    2**k phase gadgets."""
    if not check_fourier(g, v): return False

    q,r = g.qubit(v), g.row(v)
    nhd = tuple(g.neighbors(v))
    ph = g.phase(v) * Fraction(1,2**(len(nhd)-1))
    g.remove_vertex(v)
    pos = r - 0.5*(2**len(nhd) - len(nhd))

    if graph_like:
        et = EdgeType.HADAMARD
        vt = VertexType.Z
    else:
        et = EdgeType.SIMPLE
        vt = VertexType.X

    for w in nhd:
        g.add_to_phase(w, ph)
    for weight in range(2, len(nhd)+1):
        for ws in combinations(nhd, weight):
            w1 = g.add_vertex(VertexType.Z, qubit=q-1, row=pos)
            g.set_phase(w1, (-1)**(weight-1) * ph)
            w2 = g.add_vertex(vt, qubit=q, row=pos)
            g.add_edge((w1,w2), et)
            for w3 in ws:
                g.add_edge((w2, w3), et)
            g.scalar.add_power(weight-1)
            pos += 1
    return True

def check_ifourier(g: BaseGraph[VT,ET], v: VT) -> bool:
    """Returns True if the given node is the phase part of a phase gadget, and the gadget is
    connected to all Z spiders.
    
    Note this accepts phase gadgets either in graph-like form, with Hadamard edges, or in the standard
    form of a Z-spider connected to a X-spider, with normal edges."""
    ty = g.types()
    if (ty[v] != VertexType.Z or g.vertex_degree(v) != 1):
        return False
    w = next(iter(g.neighbors(v)))

    if not (
        (ty[w] == VertexType.X and all(g.edge_type(e) == EdgeType.SIMPLE for e in g.incident_edges(w))) or
        (ty[w] == VertexType.Z and all(g.edge_type(e) == EdgeType.HADAMARD for e in g.incident_edges(w)))
    ): return False

    return all(ty[w1] == VertexType.Z for w1 in g.neighbors(w))

def ifourier(g: BaseGraph[VT, ET], v: VT) -> bool:
    """Applies the inverse graphical fourier transform, replacing the given phase gadget of arity k with
    2**k H-boxes.
    
    To avoid ambiguity, the phase gadget should be specified as the arity-1 spider with the phase on it."""
    if not check_ifourier(g, v): return False

    w = next(iter(g.neighbors(v)))

    q,r = g.qubit(w), g.row(w)
    nhd = tuple(w for w in g.neighbors(w) if w != v)
    ph = g.phase(v)
    g.remove_vertex(v)
    g.remove_vertex(w)
    g.scalar.add_power(1-len(nhd))
    pos = r - 0.5*(2**len(nhd) - len(nhd))

    for w in nhd:
        g.add_to_phase(w, ph)
    for weight in range(2, len(nhd)+1):
        for ws in combinations(nhd, weight):
            w1 = g.add_vertex(VertexType.H_BOX, qubit=q, row=pos)
            g.set_phase(w1, (-2)**(weight-1) * ph)
            for w2 in ws:
                g.add_edge((w1, w2))
            pos += 1
    return True



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


usage_string = """python -m pyzx command [args]
Run one of the scripts supplied with PyZX.

The options for command are:
    opt       -- Optimize a circuit using PyZX
    tikz      -- Convert a circuit into a Tikz file
    router    -- Map any circuit onto restricted architectures
    cnots     -- Generate random CNOT circuits 
    phasepoly -- Generates random phase polynomial circuits and stores them as QASM files

For help on the arguments for these commands run for instance 'python -m pyzx opt --help'
"""

import argparse
from . import circ2circ
from . import circ2tikz
from . import circuit_router
from . import cnot_generator
from . import phase_poly_generator
   
def main(argv):
    parser = argparse.ArgumentParser(prog="PyZX", description="PyZX commandline interface",
                                     usage=usage_string)
    parser.add_argument('command', help='Command to run')
    if len(argv) == 1:
        parser.print_help()
        exit(1)
    args = parser.parse_args(argv[1:2])
    if args.command not in ('opt', 'tikz', 'router', 'cnots', 'phasepoly'):
        print("Unrecognized command '{}'".format(args.command))
        parser.print_help()
        exit(1)

    if args.command == 'opt':
        circ2circ.main(argv[2:])
    if args.command == 'tikz':
        circ2tikz.main(argv[2:])
    if args.command == 'router':
        circuit_router.main(argv[2:])
    if args.command == 'cnots':
        cnot_generator.main(argv[2:])
    if args.command == 'phasepoly':
        phase_poly_generator.main(argv[2:])


# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import math
from fractions import Fraction

from . import Circuit

def parse_quipper_block(lines: List[str]) -> Circuit:
    start = lines[0]
    end = lines[-1]
    gates = lines[1:-1]
    if not start.startswith("Inputs: "):
        raise TypeError("File does not start correctly: " + start)
    if start.endswith(','): start = start[:-1]
    inputs = start[8:].split(",")

    if inputs == ["none"]:
        inputs = []
    else:
        for ip in inputs:
            n, ty = ip.split(":")
            if ty.strip() != "Qbit":
                raise TypeError("Unsupported type " + ty)

    c = Circuit(len(inputs))
    for gate in gates:
        if gate.startswith("Comment"): continue
        if gate.startswith("QInit0"):
            t = int(gate[gate.find('(')+1:gate.find(')')])
            if t>=c.qubits: c.qubits = t+1
            continue
        if gate.startswith("QTerm0"): continue
        if gate.startswith("QMeas"): continue
        if gate.startswith("QDiscard"): continue
        if gate.startswith("QRot"):
            i = gate.find("exp(")
            if gate[i+4:i+8] == '-i%Z':
                gtype = "ZPhase"
            elif gate[i+4:i+8] == '-i%X':
                gtype = "XPhase"
            else:
                raise TypeError("Unsupported expression: " + gate) 
            val = gate[gate.find(',')+1: gate.find(']')]
            try:
                f = float(val)
            except ValueError:
                raise TypeError("Unsupported expression: " + gate)
            phase = Fraction(f/math.pi)
            target = gate[gate.rfind('(')+1:gate.rfind(')')]
            try: 
                t = int(target)
            except ValueError:
                raise TypeError("Unsupported expression: "+ gate)
            c.add_gate(gtype, t, 2*phase)
            continue
        elif not gate.startswith("QGate"):
            raise TypeError("Unsupported expression: " + gate)
        l = gate.split("with")
        g = l[0]
        gname = g[g.find('[')+2:g.find(']')-1]
        t = int(g[g.find('(')+1:g.find(')')])
        adjoint = g.find("*")!=-1
        if len(l) == 1 or (len(l) == 2 and l[1].find("nocontrol") != -1):  # no controls
            if gname == "H": c.add_gate("HAD", t)
            elif gname == "not": c.add_gate("NOT", t)
            elif gname == "Z": c.add_gate("Z", t)
            elif gname == "S": c.add_gate("S", t, adjoint=adjoint)
            elif gname == "T": c.add_gate("T", t, adjoint=adjoint)
            else:
                raise TypeError("Unsupported gate: " + gname)
            continue
        elif len(l) > 3: raise TypeError("Unsupported expression: " + gate)
        ctrls = l[1]
        ctrls = ctrls[ctrls.find('[')+1:ctrls.find(']')]
        if ctrls.find(',')!=-1:
            if ctrls.count(',') != 1: raise TypeError("Maximum two controls on gate allowed: " + gate)
            if gname not in ("not", "Z"): raise TypeError("Two controls only allowed on 'not' and 'Z': "+ gate)
            c1, c2 = ctrls.split(',',1)
            ctrl1 = int(c1.strip()[1:])
            ctrl2 = int(c2.strip()[1:])
            nots = []
            if c1.find('+') == -1:
                nots.append(ctrl1)
            if c2.find('+') == -1:
                nots.append(ctrl2)
            #if ctrl1.find('+') == -1 or ctrl2.find('+') == -1: raise TypeError("Unsupported controls: " + ctrls)
            for ts in nots: c.add_gate("NOT", ts)
            if gname == "not": c.add_gate("TOF", ctrl1, ctrl2, t)
            elif gname == "Z": c.add_gate("CCZ", ctrl1, ctrl2, t)
            for ts in nots: c.add_gate("NOT", ts)
            continue
        elif ctrls.find('+')==-1:
            raise TypeError("Unsupported control: " + ctrls)
        ctrl = int(ctrls[1:])
        if gname == "not": c.add_gate("CNOT", ctrl, t)
        elif gname == "Z": c.add_gate("CZ", ctrl, t)
        elif gname == "X": c.add_gate("CX", ctrl, t)
        else:
            raise TypeError("Unsupported controlled gate: " + gname)
    return c

def quipper_center_block(fname: str) -> Circuit:
    """Function to load the PF files of the NRSCM paper."""
    f = open(fname, 'r')
    text = f.read()
    f.close()
    i = text.find('Subroutine: "C"')
    if i == -1: 
        i = text.find('Subroutine: "S"')
        if i == -1: raise Exception("Not a valid format")
        text = text[i:].strip()
    else:
        j = text.find('Subroutine: "R"',i)
        if j == -1: raise Exception("Not a valid format")
        text = text[i:j].strip()
    lines = text.splitlines()[3:]
    return parse_quipper_block(lines)


# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
NEW VERSION 
"""

from fractions import Fraction
from itertools import combinations
from typing import Dict, List, Tuple, Callable, Optional, Set, FrozenSet
from pyzx_param.utils import EdgeType, VertexType, toggle_edge, toggle_vertex, FractionLike, FloatInt, vertex_is_zx
from pyzx_param.simplify import *
from pyzx_param.graph.base import BaseGraph, ET, VT, upair
import pyzx_param.rewrite_rules.rules as rules

def is_hadamard(g: BaseGraph[VT,ET], v: VT) -> bool:
    """Returns whether the vertex v in graph g is a Hadamard gate."""
    if g.type(v) != VertexType.H_BOX: return False
    if g.phase(v) != 1: return False
    if g.vertex_degree(v) != 2: return False
    return True

def replace_hadamard(g: BaseGraph[VT,ET], v: VT) -> bool:
    """Replaces a Hadamard gate with a Hadamard edge."""
    if not is_hadamard(g, v): return False
    n1,n2 = g.neighbors(v)
    et1 = g.edge_type(g.edge(v,n1))
    et2 = g.edge_type(g.edge(v,n2))
    if et1 == et2: # both connecting edges are HADAMARD or SIMPLE
        g.add_edge((n1,n2), EdgeType.HADAMARD)
    else:
        g.add_edge((n1,n2), EdgeType.SIMPLE)
    g.remove_vertex(v)
    g.scalar.add_power(1) # Correct for the sqrt(2) difference in H-boxes and H-edges
    return True

def had_edge_to_hbox(g: BaseGraph[VT,ET], e: ET) -> bool:
    """Converts a Hadamard edge to a Hadamard gate.
    Note that while this works with multigraphs, it will put the new H-box in the middle of the vertices,
    so that the diagram might look wrong.
    """
    et = g.edge_type(e)
    if et != EdgeType.HADAMARD: return False
    s,t = g.edge_st(e)
    rs = g.row(s)
    rt = g.row(t)
    qs = g.qubit(s)
    qt = g.qubit(t)
    g.remove_edge(e)
    h = g.add_vertex(VertexType.H_BOX)
    g.scalar.add_power(-1) # Correct for sqrt(2) scalar difference in H-edge and H-box.
    g.add_edge((s, h),EdgeType.SIMPLE)
    g.add_edge((h, t),EdgeType.SIMPLE)
    
    if qs == qt:
        g.set_qubit(h, qs)
    else:
        q = (qs + qt) / 2
        if round(q) == q: q += 0.5
        g.set_qubit(h, q)
    g.set_row(h, (rs + rt) / 2)
    return True
    

def match_hadamards(g: BaseGraph[VT,ET],
        vertexf: Optional[Callable[[VT],bool]] = None
        ) -> List[VT]:
    """Matches all the H-boxes with arity 2 and phase 1, i.e. all the Hadamard gates."""
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    m : Set[VT] = set()
    ty = g.types()
    taken: Set[VT] = set()
    for v in candidates:
        if ty[v] == VertexType.H_BOX and g.vertex_degree(v) == 2 and g.phase(v) == 1:
            n1,n2 = g.neighbors(v)
            if n1 in taken or n2 in taken: continue
            if n1 not in m and n2 not in m: 
                m.add(v)
                taken.add(n1)
                taken.add(n2)

    return list(m)

def hadamard_to_h_edge(g: BaseGraph[VT,ET], matches: List[VT]) -> rules.RewriteOutputType[VT,ET]:
    """Converts a matching of H-boxes with arity 2 and phase 1, i.e. Hadamard gates, to Hadamard edges."""
    rem_verts = []
    etab: Dict[Tuple[VT,VT], List[int]] = {}
    for v in matches:
        rem_verts.append(v)
        w1,w2 = list(g.neighbors(v))
        et1 = g.edge_type(g.edge(w1,v))
        et2 = g.edge_type(g.edge(w2,v))
        if et1 == et2:
            etab[upair(w1,w2)] = [0,1]
        else:
            etab[upair(w1,w2)] = [1,0]
    g.scalar.add_power(len(matches)) # Correct for the sqrt(2) difference in H-boxes and H-edges
    return (etab, rem_verts, [], True)

def match_connected_hboxes(g: BaseGraph[VT,ET],
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[ET]:
    """Matches Hadamard-edges that are connected to H-boxes, as these can be fused,
    see the rule (HS1) of https://arxiv.org/pdf/1805.02175.pdf. 
    
    Warning:
        Does not work with multigraphs. 
    """
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    m : Set[ET] = set()
    ty = g.types()
    while candidates:
        e = candidates.pop()
        if g.edge_type(e) != EdgeType.HADAMARD: continue
        v1,v2 = g.edge_st(e)
        if ty[v1] != VertexType.H_BOX or ty[v2] != VertexType.H_BOX: continue
        if g.phase(v1) != 1 and g.phase(v2) != 1: continue
        m.add(e)
        candidates.difference_update(g.incident_edges(v1))
        candidates.difference_update(g.incident_edges(v2))
    return list(m)

def fuse_hboxes(g: BaseGraph[VT,ET], matches: List[ET]) -> rules.RewriteOutputType[VT,ET]:
    """Fuses two neighboring H-boxes together. 
    See rule (HS1) of https://arxiv.org/pdf/1805.02175.pdf."""
    rem_verts = []
    etab: Dict[Tuple[VT,VT], List[int]] = {}
    for e in matches:
        v1, v2 = g.edge_st(e)
        if g.phase(v2) != 1: # at most one of v1 and v2 has a phase different from 1
            v1, v2 = v2, v1
        rem_verts.append(v2)
        g.scalar.add_power(1)
        for n in g.neighbors(v2):
            if n == v1: continue
            e2 = g.edge(v2,n)
            if g.edge_type(e2) == EdgeType.SIMPLE:
                etab[upair(v1,n)] = [1,0]
            else:
                etab[upair(v1,n)] = [0,1]

    return (etab, rem_verts, [], True)



MatchCopyType = Tuple[VT,VT,VertexType,FractionLike,FractionLike,List[ET]]

def match_copy(
        g: BaseGraph[VT,ET], 
        vertexf:Optional[Callable[[VT],bool]]=None
        ) -> List[MatchCopyType[VT, ET]]:
    """Finds arity-1 spiders (with a 0 or pi phase) that can be copied through their neighbor.""" 
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    phases = g.phases()
    types = g.types()
    m = []
    taken: Set[VT] = set()

    while len(candidates) > 0:
        v = candidates.pop()
        if phases[v] not in (0,1) or types[v] == VertexType.BOUNDARY or g.vertex_degree(v) != 1:
                    continue
        w = list(g.neighbors(v))[0]
        if w in taken: continue
        tv = types[v]
        if tv == VertexType.H_BOX: tv = VertexType.Z # v is arity 1, so we can treat it as a Z spider
        tw = types[w]
        if tw == VertexType.BOUNDARY: continue
        e = g.edge(v,w)
        et = g.edge_type(e)
        copy_type: VertexType = VertexType.Z
        if vertex_is_zx(tv):
            if vertex_is_zx(tw):
                if et == EdgeType.HADAMARD:
                    if tw != tv: continue
                    copy_type = toggle_vertex(tv)
                else:
                    if tw == tv: continue
                    copy_type = tv
            elif tw == VertexType.H_BOX:
                # X pi/0 can always copy through H-box
                # But if v is Z, then it can only copy if the phase is 1
                if et == EdgeType.HADAMARD:
                    if tv == VertexType.Z:
                        if phases[v] == 1: copy_type = VertexType.BOUNDARY # We don't actually copy in this case
                        else: copy_type = VertexType.Z
                    else:
                        if phases[v] != 1: continue
                        copy_type = VertexType.X
                else:
                    if tv == VertexType.X:
                        if phases[v] == 1: copy_type = VertexType.BOUNDARY # We don't actually copy in this case
                        else: copy_type = VertexType.Z
                    else:
                        if phases[v] != 1: continue
                        copy_type = VertexType.X
            else:
                continue
        else:
            continue
        neigh = [n for n in g.neighbors(w) if n != v]
        neigh_edges = [e for e in g.incident_edges(w) if v not in g.edge_st(e)]
        m.append((v,w,copy_type,phases[v],phases[w],neigh_edges))
        candidates.discard(w)
        candidates.difference_update(neigh)
        taken.add(w)
        taken.update(neigh)

    return m

def apply_copy(
        g: BaseGraph[VT,ET],
        matches: List[MatchCopyType[VT, ET]]
        ) -> rules.RewriteOutputType[VT,ET]:
    """Copy arity-1 spider through their neighbor."""
    rem = []
    types = g.types()
    for v,w,copy_type,a,alpha,neigh_edges in matches:
        rem.append(v)
        if copy_type == VertexType.BOUNDARY:
            g.scalar.add_power(1)
            continue # Don't have to do anything more for this case
        rem.append(w)
        if vertex_is_zx(types[w]):
            if a: g.scalar.add_phase(alpha)
            g.scalar.add_power(-(len(neigh_edges)-1))
        else: #types[w] == H_BOX
            if copy_type == VertexType.Z:
                g.scalar.add_power(1)
            else:
                g.scalar.add_power(-(len(neigh_edges)-2))
                if alpha != 1:
                    g.scalar.add_power(-2)
                    g.scalar.add_node(alpha+1)
        for edge in neigh_edges:
            st = g.edge_st(edge)
            n = st[0] if st[1] == w else st[1]
            r = 0.7*g.row(w) + 0.3*g.row(n)
            q = 0.7*g.qubit(w) + 0.3*g.qubit(n)

            u = g.add_vertex(copy_type, q, r, a)
            et = g.edge_type(edge)
            g.add_edge((n,u), et)

    return ({}, rem, [], True)


def is_NOT_gate(g, v, n1, n2):
    """Returns whether the vertex v in graph g is a NOT gate between its neighbours n1 and n2."""
    return (
        (
            g.edge_type(g.edge(n1,v)) == EdgeType.SIMPLE 
            and g.type(v) == VertexType.X 
            and g.edge_type(g.edge(n2,v)) == EdgeType.SIMPLE
        ) or (
            g.edge_type(g.edge(n1,v)) == EdgeType.HADAMARD 
            and g.type(v) == VertexType.Z 
            and g.edge_type(g.edge(n2,v)) == EdgeType.HADAMARD
        )
    ) 


def match_hbox_parallel_not(
        g: BaseGraph[VT,ET], 
        vertexf:Optional[Callable[[VT],bool]]=None
        ) -> List[Tuple[VT,VT,VT]]:
    """Finds H-boxes that are connected to a Z-spider both directly and via a NOT.""" 
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    phases = g.phases()
    types = g.types()
    m = []

    while len(candidates) > 0:
        h = candidates.pop()
        if types[h] != VertexType.H_BOX or phases[h] != 1: continue

        for n in g.neighbors(h):
            if g.vertex_degree(n) != 2 or phases[n] != 1: continue # If it turns out to be useful, this rule can be generalised to allow spiders of arbitrary phase here
            v = [v for v in g.neighbors(n) if v != h][0] # The other neighbor of n
            if not g.connected(v,h): continue
            if not is_NOT_gate(g,n,h,v):
                continue
            break
        else:
            continue
        # h is connected to both v and n in the appropriate way, and n is a NOT that is connected to v as well
        m.append((h,v,n))
        candidates.difference_update(g.neighbors(h))
    return m

def hbox_parallel_not_remove(g: BaseGraph[VT,ET], 
        matches: List[Tuple[VT,VT,VT]]
        ) -> rules.RewriteOutputType[VT,ET]:
    """If a Z-spider is connected to an H-box via a regular wire and a NOT, then they disconnect, and the H-box is turned into a Z-spider."""
    rem = []
    etab: Dict[Tuple[VT,VT], List[int]] = {}
    types = g.types()
    for h, v, n in matches:
        rem.append(h)
        rem.append(n)
        for w in g.neighbors(h):
            if w == v or w == n: continue
            et = g.edge_type(g.edge(w,h))
            if types[w] == VertexType.Z and et == EdgeType.SIMPLE: continue
            if types[w] == VertexType.X and et == EdgeType.HADAMARD: continue
            q = 0.6*g.qubit(h) + 0.4*g.qubit(w)
            r = 0.6*g.row(h) + 0.4*g.row(w)
            z = g.add_vertex(VertexType.Z,q,r)
            if et == EdgeType.SIMPLE:
                etab[upair(z,w)] = [1,0]
            else: etab[upair(z,w)] = [0,1]
    return (etab, rem, [], True)


TYPE_MATCH_PAR_HBOX = Tuple[List[VT],List[VT],List[VT]]
def match_par_hbox(
    g: BaseGraph[VT,ET],
    vertexf: Optional[Callable[[VT],bool]] = None
    ) -> List[TYPE_MATCH_PAR_HBOX]:
    """Matches sets of H-boxes that are connected in parallel (via optional NOT gates)
    to the same white spiders."""
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    
    groupings: Dict[Tuple[FrozenSet[VT],FrozenSet[VT]], Tuple[List[VT],List[VT],List[VT]]] = dict()
    ty = g.types()
    for h in candidates:
        if ty[h] != VertexType.H_BOX: continue
        suitable = True
        neighbors_regular = set()
        neighbors_NOT = set()
        NOTs = []
        for v in g.neighbors(h):
            e = g.edge(v,h)
            if g.edge_type(e) == EdgeType.HADAMARD:
                if ty[v] == VertexType.X:
                    neighbors_regular.add(v)
                elif ty[v] == VertexType.Z and g.vertex_degree(v) == 2 and g.phase(v) == 1:
                    w = [w for w in g.neighbors(v) if w!=h][0] # unique other neighbor
                    if ty[w] != VertexType.Z or g.edge_type(g.edge(v,w)) != EdgeType.HADAMARD:
                        suitable = False
                        break
                    neighbors_NOT.add(w)
                    NOTs.append(v)
                else:    
                    suitable = False
                    break
            else: # e == EdgeType.SIMPLE
                if ty[v] == VertexType.Z:
                    neighbors_regular.add(v)
                elif ty[v] == VertexType.X and g.vertex_degree(v) == 2 and g.phase(v) == 1:
                    w = [w for w in g.neighbors(v) if w!=h][0] # unique other neighbor
                    if ty[w] != VertexType.Z or g.edge_type(g.edge(v,w)) != EdgeType.SIMPLE:
                        suitable = False
                        break
                    neighbors_NOT.add(w)
                    NOTs.append(v)
                else:
                    suitable = False
                    break
        if not suitable: continue
        group = (frozenset(neighbors_regular), frozenset(neighbors_NOT))
        if group in groupings: 
            groupings[group][0].append(h)
            groupings[group][2].extend(NOTs)
        else: groupings[group] = ([h],NOTs, [])

    m = []
    for (n_r, n_N), (hs,firstNOTs, NOTs) in groupings.items():
        if len(hs) < 2: continue
        m.append((hs, firstNOTs, NOTs))
    return m

def par_hbox(g: BaseGraph[VT,ET], matches: List[TYPE_MATCH_PAR_HBOX]) -> rules.RewriteOutputType[VT,ET]:
    """Implements the `multiply rule' (M) from https://arxiv.org/abs/1805.02175"""
    rem_verts = []
    for hs, firstNOTs, NOTs in matches:
        p = sum(g.phase(h) for h in hs) % 2
        rem_verts.extend(hs[1:])
        rem_verts.extend(NOTs)
        if p == 0: 
            rem_verts.append(hs[0])
            rem_verts.extend(firstNOTs)
        else: g.set_phase(hs[0], p)
    
    return ({}, rem_verts, [], False)


TYPE_MATCH_PAR_HBOX_INTRO = Tuple[VT,VT,VT,List[VT],Set[VT]]
def match_par_hbox_intro(
    g: BaseGraph[VT,ET],
    vertexf: Optional[Callable[[VT],bool]] = None
    ) -> List[TYPE_MATCH_PAR_HBOX_INTRO]:
    """Matches sets of H-boxes that are connected in parallel (via optional NOT gates)
    to the same white spiders, but with just one NOT different, so that the Intro rule can be applied there."""
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    
    groupings: Dict[FrozenSet[VT], List[Tuple[VT,List[VT],Set[VT],Set[VT],Set[VT]]]] = dict()
    ty = g.types()
    for h in candidates:
        if ty[h] != VertexType.H_BOX: continue
        suitable = True
        neighbors_regular = set()
        neighbors_NOT = set()
        neighbors_single = set()  # Single-arity Z-spiders connected to the H-box.
        NOTs = []
        for v in g.neighbors(h):
            e = g.edge(v,h)
            if g.edge_type(e) == EdgeType.HADAMARD:
                if ty[v] != VertexType.Z or g.vertex_degree(v) != 2 or g.phase(v) != 1: 
                    suitable = False
                    break
                w = [w for w in g.neighbors(v) if w!=h][0]  # unique other neighbor
                if ty[w] != VertexType.Z or g.edge_type(g.edge(v,w)) != EdgeType.HADAMARD:
                    suitable = False
                    break
                neighbors_NOT.add(w)
                NOTs.append(v)
            else: # e == EdgeType.SIMPLE
                if ty[v] == VertexType.Z:
                    if g.vertex_degree(v) == 1:
                        if g.phase(v) != 0:
                            suitable = False
                            break
                        neighbors_single.add(v)
                    else:
                        neighbors_regular.add(v)
                else:
                    if ty[v] != VertexType.X or g.vertex_degree(v) != 2 or g.phase(v) != 1:
                        suitable = False
                        break
                    w = [w for w in g.neighbors(v) if w!=h][0]  # unique other neighbor
                    if ty[w] != VertexType.Z or g.edge_type(g.edge(v,w)) != EdgeType.SIMPLE:
                        suitable = False
                        break
                    neighbors_NOT.add(w)
                    NOTs.append(v)
        if not suitable: continue
        group = frozenset(neighbors_regular | neighbors_NOT)
        if group not in groupings: 
            groupings[group] = [(h,NOTs, neighbors_regular, neighbors_NOT, neighbors_single)]
            continue
        # There is another H-box with the same set of neighbours
        for h2, NOTs2, neighbors_regular2, neighbors_NOT2, neighbors_single2 in groupings[group]:
            vs = neighbors_regular.symmetric_difference(neighbors_regular2)
            if g.phase(h) != g.phase(h2): continue  # TODO: Allow intro rule with different phases.
            if len(neighbors_single) != len(neighbors_single2): continue  # TODO: Allow different sets of neighbours here.
            if len(vs) != 1: # To use the Intro rule, the H-boxes should differ on exactly one position on where there are NOTs.
                continue
            # We have a match!
            v = vs.pop() 
            break
        else:
            continue
        if v in neighbors_regular2:  # Make it so that h is the vertex that remains
            h,h2 = h2, h
            NOTs2 = NOTs
            neighbors_single2 = neighbors_single
        return [(h,h2,v,NOTs2,neighbors_single2)]
    return []

def par_hbox_intro(g: BaseGraph[VT,ET], matches: List[TYPE_MATCH_PAR_HBOX_INTRO]) -> rules.RewriteOutputType[VT,ET]:
    """Removes an H-box according to the Intro rule (See Section 3.2 of arxiv:2103.06610)."""
    rem_verts = []
    rem_edges = []
    for h, h2, v, NOTs, singles in matches:
        rem_verts.append(h2)
        rem_verts.extend(singles)
        rem_verts.extend(NOTs)
        rem_edges.append(g.edge(h,v))
        g.scalar.add_power(2*len(singles))
    return ({}, rem_verts, rem_edges, True)


def match_zero_hbox(g: BaseGraph[VT,ET]) -> List[VT]:
    """Matches H-boxes that have a phase of 2pi==0."""
    types = g.types()
    phases = g.phases()
    return [v for v in g.vertices() if types[v] == VertexType.H_BOX and phases[v] == 0]

def zero_hbox(g: BaseGraph[VT,ET], m: List[VT]) -> None:
    """Removes H-boxes with a phase of 2pi=0.
    Note that this rule is only semantically correct when all its neighbors are white spiders."""
    g.remove_vertices(m)


hpivot_match_output = List[Tuple[
            VT,
            VT,
            VT,
            List[VT],
            List[VT],
            List[List[VT]],
            List[Tuple[FractionLike,List[VT]]]
            ]]

def match_hpivot(
    g: BaseGraph[VT,ET], matchf=None
    ) -> hpivot_match_output:
    """Finds a matching of the hyper-pivot rule. Note this currently assumes
    hboxes don't have phases.

    :param g: An instance of a ZH-graph.
    :param matchf: An optional filtering function for candidate arity-2 hbox, should
       return True if an hbox should considered as a match. Passing None will
       consider all arity-2 hboxes.
    :rtype: List containing 0 or 1 matches.
    """

    types = g.types()
    phases = g.phases()
    m = []

    min_degree = -1

    for h in g.vertices():
        if not (
            (matchf is None or matchf(h)) and
            g.vertex_degree(h) == 2 and
            types[h] == VertexType.H_BOX and
            phases[h] == 1
        ): continue

        v0, v1 = g.neighbors(h)

        v0n = set(g.neighbors(v0))
        v1n = set(g.neighbors(v1))

        if (len(v0n.intersection(v1n)) > 1): continue

        v0b = [v for v in v0n if types[v] == VertexType.BOUNDARY]
        v0h = [v for v in v0n if types[v] == VertexType.H_BOX and v != h]
        v1b = [v for v in v1n if types[v] == VertexType.BOUNDARY]
        v1h = [v for v in v1n if types[v] == VertexType.H_BOX and v != h]

        # check that at least one of v0 or v1 has all pi phases on adjacent
        # hboxes.
        if not (all(phases[v] == 1 for v in v0h)):
            if not (all(phases[v] == 1 for v in v1h)):
                continue
            else:
                # interchange the roles of v0 <-> v1
                v0,v1 = v1,v0
                v0n,v1n = v1n,v0n
                v0b,v1b = v1b,v0b
                v0h,v1h = v1h,v0h

        v0nn = [list(filter(lambda w : w != v0, g.neighbors(v))) for v in v0h]
        v1nn = [
          (phases[v],
           list(filter(lambda w : w != v1, g.neighbors(v))))
          for v in v1h]


        if not (
            all(all(types[v] == VertexType.Z for v in vs) for vs in v0nn) and
            all(all(types[v] == VertexType.Z for v in vs[1]) for vs in v1nn) and
            len(v0b) + len(v1b) <= 1 and
            len(v0b) + len(v0h) + 1 == len(v0n) and
            len(v1b) + len(v1h) + 1 == len(v1n)
        ): continue

        degree = g.vertex_degree(v0) * g.vertex_degree(v1)

        if min_degree == -1 or degree < min_degree:
            m = [(h, v0, v1, v0b, v1b, v0nn, v1nn)]
            min_degree = degree
    return m


def hpivot(g: BaseGraph[VT,ET], m: hpivot_match_output) -> None:
    if len(m) == 0: return None

    types = g.types()

    # # cache hboxes
    # hboxes = dict()
    # for h in g.vertices():
    #     if types[h] != VertexType.H_BOX: continue
    #     nhd = tuple(sorted(g.neighbors(h)))
    #     hboxes[nhd] = h


    h, v0, v1, v0b, v1b, v0nn, v1nn = m[0]
    g.remove_vertices([v for v in g.neighbors(v0) if types[v] == VertexType.H_BOX])
    g.remove_vertices([v for v in g.neighbors(v1) if types[v] == VertexType.H_BOX])
    g.scalar.add_power(2) # Applying a Fourier Hyperpivot adds a scalar of 2
    
    if len(v0b) == 0:
        g.remove_vertex(v0)
    else:
        e = g.edge(v0, v0b[0])
        g.set_edge_type(e, toggle_edge(g.edge_type(e)))
        v0nn.append([v0])
    
    if len(v1b) == 0:
        g.remove_vertex(v1)
    else:
        e = g.edge(v1, v1b[0])
        g.set_edge_type(e, toggle_edge(g.edge_type(e)))
        v1nn.append((Fraction(1,1), [v1]))

    for phase,ws in v1nn:
        for weight in range(1,len(v0nn)+1):
            phase_mult = int((-2)**(weight-1))
            f_phase = (phase * phase_mult) % 2
            if f_phase == 0: continue
            for vvs in combinations(v0nn, weight):
                us = tuple(sorted(sum(vvs, ws)))

                # TODO: check if this is the right thing to do (and update scalar)
                if len(us) == 0: continue

                # if us in hboxes:
                #     h0 = hboxes[us]
                #     print("adding %s to %s" % (f_phase, g.phase(h0)))
                #     g.add_to_phase(h0, f_phase)
                # else:
                h0 = g.add_vertex(VertexType.H_BOX)
                g.set_phase(h0, f_phase)
                q: FloatInt = 0
                r: FloatInt = 0
                for u in us:
                    q += g.qubit(u)
                    r += g.row(u)
                    g.add_edge((h0,u))
                g.set_qubit(h0, q / len(us) - 0.4)
                g.set_row(h0, r / len(us) + 0.4)